;;; -*- coding: utf-8 -*-
;;; Check whether a file contains expressions within the LI101 dialect.
;;; then evaluate an LI101 program.

;;; {{{ Syntax check
;;; All errors are signaled through li101-error.
;;; check-* function calls li101-error in case of problems
;;; is-* are real predicates.

;;; This variable is an Alist of alternatives. It will be enriched by
;;; the evaluator with the number of times the then or else branches
;;; are chosen.
(define *alternatives* (list))

;;; slurp-file: FileName -> List(Expression) + ERREUR
(define (slurp-file file)
  (with-input-from-file file
    (lambda ()
      (let loop ()
        ;(display `(about to read))(newline);DEBUG
        (let ((e (read)))
          ;(display `(read ,e))(newline); DEBUG
          (if (eof-object? e)
              (list)
              (cons e (loop)) ) ) ) ) ) )

;;; check-program?: List(Expression) -> Bool + ERREUR
;;; Most of the time check-program? raises an error when a bad syntax is 
;;; discovered however sometime it just returns false (TO BE FIXED).
;;; program := { define | verify | expression }+

(define (check-program? e*)
  (if (is-non-empty-list? e*)
      (check-toplevel-expressions e*)
      (li101-error 1102 "Programme vide") ) )

(define (check-toplevel-expressions e*)
  (if (pair? e*)
      (and (cond ((is-define? (car e*))
                  (check-define? (car e*)) )
                 ((is-verify? (car e*))
                  (check-verify? (car e*)) )
                 ;; For teachers
                 ((is-li101-verify? (car e*))
                  (check-li101-verify? (car e*)) )
                 ;; Default:
                 (else (check-expression? (car e*))) )
           (check-toplevel-expressions (cdr e*)) )
      (if (null? e*)
          #t
          (li101-error 1142 "Fin de programme incorrecte" e*) ) ) )

(define (is-define? e)
  (and (pair? e)
       (eq? 'define (car e)) ) )
(define (is-define-constant? e)
  (and (is-define? e)
       (pair? (cdr e))
       (not (pair? (cadr e))) ) )
(define (is-define-function? e)
  (and (is-define? e)
       (pair? (cdr e))
       (pair? (cadr e)) ) )

;;; (define var exp) | (define (var var...) exp+)
(define (check-define? e)
  (if (is-define? e)
      (if (pair? (cdr e))
          (if (pair? (cadr e))
              (check-define-fonction? e)
              (check-define-constant? e) )
          (li101-error 1103 "Définition tronquée" e) )
      (li101-error 1104 "J'attendais une définition" e) ) )

;;; (define var exp)
(define (check-define-constant? e)
  ;; (assume (and (pair? e) (pair? (cdr e))))
  (if (is-variable? (cadr e))
      (if (and (pair? (cddr e))
               (check-expression? (caddr e))
               (null? (cdddr e)) )
          #t
          (li101-error 1105 "Mauvaise définition de constante" e) )
      (li101-error 1106 "Définition incorrecte de constante" e) ) )

;;; (define (var var...) definition* exp+)
(define (check-define-fonction? e)
  ;; (assume (and (pair? e) (pair? (cdr e)) (pair? (cadr e))))
  (if (and (is-non-empty-list? (cadr e))
           (all? is-variable? (cadr e))
           (check-define-body (cddr e)) )
      #t
      (li101-error 1107 "Mauvaise définition de fonction" e) ) )

;;; (define ..)* expression+
;;; Used in (define) and (let|let*) forms.
(define (check-define-body e*)
  (if (pair? e*)
      (if (is-define? (car e*))
          (and (check-define? (car e*))
               (check-define-body (cdr e*)) )
          (and (is-non-empty-list? e*)
               (all? check-expression? e*) ) )
      (li101-error 1108 "J'attends un corps d'au moins une expression" e*) ) )

(define (is-verify? e)
  (and (pair? e)
       (memq (car e) '(verifier verify tester)) ) )

;;; (verifier nom { exp => Sexp }+)
;;; (verifier (nom+) { exp => Sexp }+)
(define (check-verify? e)
  (if (is-verify? e)
      (if (and (pair? (cdr e))
               (or (is-variable? (cadr e))
                   (and (pair? (cadr e))
                        (andmap is-variable? (cadr e)) ) ) )
          (check-verify-clause+ (cddr e))
          (li101-error 1109 "Vérification incomplète" e) )
      (li101-error 1110 "J'attendais une vérification" e) ) )

;;; { exp => Sexp }+ but at least two such tests.
(define (check-verify-clause+ ee*)
  (define (check-verify-clause e* count)
    (if (pair? e*)
        (if (and (check-expression? (car e*))
                 (pair? (cdr e*))
                 (test-symbol? (cadr e*))
                 (pair? (cddr e*)) )
            (and (check-s-expression? (caddr e*))
                 (check-verify-clause (cdddr e*) (+ 1 count)) )
            (li101-error 1112 "test incorrect" 
                         (list (car e*)(cadr e*)(caddr e*)) ) )
        (cond ((null? e*)
               (if (< count 2)
                   (li101-error 1112 "J'attends au moins deux tests" ee*)
                   #t ) )
              (else
               (li101-error 1112 "Je n'attendais pas cela" e*) ) ) ) )
  (if (pair? ee*)
      (check-verify-clause ee* 0)
      (li101-error 1113 "J'attends au moins un test" ee*) ) )

(define *test-symbols*
  ;; deprecated: ~> 
  '(=> -> ~>) )

(define (test-symbol? e)
  (and (symbol? e)
       (memq e *test-symbols*) ) )

;;; (li101-verify (functions1...) (functions2...) int => exp ...)
;;; functions1 are the functions to profile. functions2 are the
;;; functions to be captured in the student or teacher environments.
(define (check-li101-verify? e)
  (define (check-li101-verify-clauses e*)
    (cond ((pair? e*)
           (if (and (integer? (car e*))
                    (pair? (cdr e*))
                    (test-symbol? (cadr e*))
                    (pair? (cddr e*)) )
               (and (check-expression? (caddr e*))
                    (check-li101-verify-clauses (cdddr e*)) )
               (li101-error 1112 "Incorrect profiling" 
                            (list (car e*)(cadr e*)(caddr e*)) ) ) )
          ((null? e*) #t)
          (else (li101-error 1112 "Bad profiling clause" e*)) ) )
  (if (is-li101-verify? e)
      (if (and (pair? (cdr e))
               (and (pair? (cadr e))
                    (andmap is-variable? (cadr e)) ) )
          (if (and (pair? (caddr e))
                   (andmap is-variable? (caddr e)) )
              (check-li101-verify-clauses (cdddr e))
              (li101-error 1109 "Incorrect function names" e) )
          (li101-error 1109 "Incorrect profiled functions names" e) )
      (li101-error 1110 "Incorrect form" e) ) )

(define (is-li101-verify? e)
  (and (pair? e)
       (memq (car e) '(li101-verify li101-verifier)) ) )

(define *allowed-keywords* 
  '(quote if cond begin and or let let*) )
(define (allow-keyword! kw)
  (set! *allowed-keywords* (cons kw *allowed-keywords*)) )
(define (disallow-keyword! kw)
  (set! *allowed-keywords* (remq kw *allowed-keywords*)) )
(define (is-allowed-keyword? kw)
  (memq kw *allowed-keywords*) )

(define (check-expression? e)
  (define errmsg "Le mot clé suivant n'est pas permis")
  (if (pair? e)
      (case (car e)
        ((quote) 
         (if (is-allowed-keyword? (car e))
             (check-quote (cdr e))
             (li101-error 1160 errmsg (car e)) ) )
        ((if)    
         (if (is-allowed-keyword? (car e))
             (check-if e)
             (li101-error 1160 errmsg (car e)) ) )
        ((cond)  
         (if (is-allowed-keyword? (car e))
             (check-cond (cdr e))
             (li101-error 1160 errmsg (car e)) ) )
        ((begin) 
         (if (is-allowed-keyword? (car e))
             (check-begin (cdr e))
             (li101-error 1160 errmsg (car e)) ) )
        ((and or) 
         (if (is-allowed-keyword? (car e))
             (check-and-or (cdr e))
             (li101-error 1160 errmsg (car e)) ) )
        ((let let*) 
         (if (is-allowed-keyword? (car e))
             (check-let (cdr e))
             (li101-error 1160 errmsg (car e)) ) )
        ;; lambda n'est pas permis aux etudiants mais permis pour l'interprete!
        (else    (check-application e)) )
      (check-atom e) ) )

(define (check-and-or e*)
  (all? check-expression? e*) )
(define (check-let e)
  (define (check-bindings e*)
    (if (pair? e*)
        (and (check-binding (car e*))
             (check-bindings (cdr e*)) )
        (null? e*) ) )
  (define (check-binding e)
    (if (and (is-non-empty-list? e)
             (is-variable? (car e))
             (pair? (cdr e))
             (check-expression? (cadr e))
             (null? (cddr e)) )
        #t
        (li101-error 1114 "Liaison incorrecte" e) ) )
  (if (and (pair? e)
           (check-bindings (car e)) )
      (check-define-body (cdr e))
      (li101-error 1115 "Variables locales mal formées" e) ) )

(define (check-atom e)
  (cond 
   ((or (number? e)(string? e)(boolean? e)(is-variable? e))
    #t )
   ((equal? e '())
    (li101-error 1116 "Vouliez-vous écrire (list) plutôt que" e) )
   ((member e *forbidden-symbols*)
    (li101-error 1116 "Ce ne peut être le nom d'une variable" e) )
   (else (li101-error 1116 "Incorrect" e)) ) )

(define (check-application e)
  (if (and (is-non-empty-list? e)
           (all? check-expression? e)
           (not (number? (car e)))
           (not (string? (car e)))
           (not (boolean? (car e))) )
      #t
      (li101-error 1117 "Application incorrecte" e) ) )

;;; (quote sexp)
(define (check-quote e*)
  (if (and (pair? e*)
           (check-s-expression? (car e*))
           (null? (cdr e*)) )
      #t
      (li101-error 1118 "Citation incorrecte" e*) ) )  

;;; (if exp exp exp?)
(define (check-if e)
  (let ((e* (cdr e)))
    (if (and (pair? e*)
             (check-expression? (car e*))
             (pair? (cdr e*))
             (check-expression? (cadr e*))
             (or (and (pair? (cddr e*))
                      (check-expression? (caddr e*))
                      (null? (cdddr e*)) )
                 ;; Ajoutons une clause else #f:
                 (if (null? (cddr e*))
                     (begin (set-cdr! (cdr e*) (list #f))
                            #t )
                     #f ) ) )
        (begin 
          (set! *alternatives*
                (cons (list e 0 0) *alternatives*) )
          #t )
        (li101-error 1119 "Alternative incorrecte" `(if . ,e*)) ) ) )

;;; (cond (exp exp)+ )
;;; NOTA: on permet (cond (exp exp+)+) ???
(define (check-cond e*)
  (define (check-cond-clause e)
    (if (and (pair? e)
             ;; else ne peut etre que la derniere clause!
             (not (eq? (car e) 'else))
             (check-expression? (car e))
             (pair? (cdr e))
             (is-non-empty-list? (cdr e))
             (all? check-expression? (cdr e)) )
        #t
        (li101-error 1120 "clause incorrecte" e) ) )
  (define (check-cond-else-clause e)
    (and (pair? e)
         (eq? (car e) 'else)
         (pair? (cdr e))
         (is-non-empty-list? (cdr e))
         (all? check-expression? (cdr e)) ) )
  (if (and (is-non-empty-list? e*)
           (all-but-last? check-cond-clause e*) )
      (let ((lastclauses (tail e*)))
        (or (check-cond-else-clause (car lastclauses))
            (begin 
              (set-cdr! lastclauses `((else #f)))
              #t ) ) )
      (li101-error 1121 "Forme cond incorrecte" `(cond . ,e*)) ) )

;;; (begin e+)
(define (check-begin e*)
  (if (and (is-non-empty-list? e*)
           (all? check-expression? e*) )
      #t
      (li101-error 1122 "Séquence incorrecte" `(begin . ,e*)) ) )

(define (check-s-expression? e)
  (if (or (symbol? e)
          (number? e)
          (string? e)
          (boolean? e)
          (null? e)
          (and (is-non-empty-list? e)
               (all? check-s-expression? e) ) )
      #t
      (li101-error 1123 "Constante bizarroïde" e) ) )

(define (all? p e*)
  (or (null? e*)
      (and (pair? e*)
           (p (car e*))
           (all? p (cdr e*)) ) ) )

(define (all-but-last? p e*)
  (if (pair? e*)
      (or (null? (cdr e*))
          (and (p (car e*))
               (all-but-last? p (cdr e*)) ) )
      #t ) )

(define (last e*)
  (define (fin e*)
    ;; assume (pair? e*)
    (if (pair? (cdr e*))
        (fin (cdr e*))
        (car e*) ) )
  (if (pair? e*)
      (fin e*)
      (li101-error 1143 "internal error: (last not-pair) !?") ) )

(define (tail e*)
  (define (fin e*)
    ;; assume (pair? e*)
    (if (pair? (cdr e*))
        (fin (cdr e*))
        e* ) )
  (if (pair? e*)
      (fin e*)
      (li101-error 1144 "internal error: (tail not-pair) !?") ) )

(define (is-non-empty-list? e)
  (define (possibly-empty-list? e)
    (if (pair? e)
        (possibly-empty-list? (cdr e))
        (null? e) ) )
  (if (pair? e)
      (possibly-empty-list? e)
      #f ) )

(define *forbidden-symbols* 
  '( if define 
     verifier tester 
     begin cond quote and or else let let* lambda ) )

(define (is-variable? e)
  (and (symbol? e)
       (not (memq e *forbidden-symbols*)) ) )

;;; }}}
;;; {{{ Evaluation
;;; eval : program -> ( (name function checker) ...)
;;; function is the value of name and checker comes from the associated
;;; (verifier name ...) form. The signature of checker is (lambda (name) ...)
;;; Program is a list { (define ...)+ (verifier ...)+ }+

;;; Enregistrer l'usage de mots ou fonctions nommees:
;;; Hashtable = FunctionName -> Number
(define hprofiling (make-hashtable))
(define (reset-profiling! . h)
  (let ((oldh hprofiling))
    (set! hprofiling (if (pair? h) (car h) (make-hashtable)))
    oldh ) )
(define (incr-profiling name . increment)
  (let ((increment (if (pair? increment) (car increment) 1))
        (n (or (hashtable-get hprofiling name) 0)) )
    (debug `(incr-profiling ,name with ,n)) ; DEBUG
    (hashtable-put! hprofiling name (+ n increment)) ) )
(define (get-profiling name)
  (let ((n (or (hashtable-get hprofiling name) 0)))
    n ) )    
(define (make-profiling name f)
  (debug `(make-profiling ,name with ,f)) ; DEBUG
  (lambda args
    (incr-profiling name)
    (apply f args) ) )
;;; v is teacher-v or student-v
(define (extract-profiling v)
  (let loop ((v v)
             (records (list)) )
    (if (pair? v)
        (if (eq? (caar v) adjoined-complexity-tag)
            (loop (cdr v) (cons (cdar v) records))
            (loop (cdr v) records) )
        (sort (lambda (r1 r2)
                (< (car r1) (car r2)) )
              records ) ) ) )


;;; the epsilon for the comparison of float numbers:
(define *epsilon* 0.01)

;;; In order to distinguish between #f (absence of value) and #f as a
;;; value, all values are put in a list.
(define (global-hashtable-put! hashtable name value)
  (hashtable-put! hashtable name (list value)) )

;;; Verification d'arites
(define (check-arity? name function arity comparator)
  (lambda args
    (let ((arglength (length args)))
      (if (comparator arity arglength)
          (apply function args)
          (li101-error "Nombre d'arguments incorrect pour" name) ) ) ) )

;;; Verification d'arite pour operateur tel que - (unaire ou binaire):
(define (or-one-more? n arglength)
  (or (= n arglength)
      (= (+ n 1) arglength) ) )

;;; egalite avec traitement des nombres flottants:
(define (quasi-equal? a b)
  (define (atom? a) (not (pair? a)))
  (cond ((and (number? a) (number? b))
         (<= (abs (- a b)) *epsilon*) )
        ((and (atom? a) (atom? b))
         (equal? a b) )
        ((and (pair? a) (pair? b))
         (and (quasi-equal? (car a) (car b))
              (quasi-equal? (cdr a) (cdr b)) ) )
        (else #f ) ) )

;;; Verification de type et qualité:
(define (check-numbers? name f)
  (lambda args
    (if (andmap number? args)
        (apply f args)
        (li101-error "un des arguments de " name
                     "n'est pas un nombre!" args) ) ) )
(define (check-division-numbers? name f)
  (lambda args
    (if (andmap number? args)
        (if (and (pair? args)
                 (pair? (cdr args))
                 (not (= 0 (cadr args))) )
            (apply f args)
            (li101-error "Division par zéro!" args) )
        (li101-error "un des arguments de " name
                     "n'est pas un nombre!" args) ) ) )
(define (check-integers? name f)
  (lambda args
    (if (andmap integer? args)
        (apply f args)
        (li101-error "un des arguments de " name
                     "n'est pas un entier!" args) ) ) )
(define (check-pair? name f)
  (lambda (args)
    (if (and (pair? args)
             (andmap pair? args) )
        (apply f args)
        (li101-error "le premier argument de" name 
                     "ne vérifie pas pair?" (car args) ) ) ) )
(define (check-positive-integers? name f)
  (lambda args
    (and (check-integers? name args)
         (if (andmap positive? args)
             (apply f args)
             (li101-error "un des arguments de " name
                          "n'est pas un entier naturel!" args) ) ) ) )
(define (check-non-negative-numbers? name f)
  (lambda args
    (define (positive? x)
      (>= x 0) )
    (and (check-numbers? name args)
         (if (andmap positive? args)
             (apply f args)
             (li101-error "un des arguments de " name
                          "n'est pas un nombre positif!" args) ) ) ) )

(define (build-primitive-environment-arithmetic)
  (let ((h (make-hashtable)))
    (define (insert name function arity comparator)
      (let ((f (check-arity? name function arity comparator)))
        (global-hashtable-put! h name f) ) )
    (insert 'equal? quasi-equal? 2 =)
    (insert 'not not 1 =)
    (insert 'number? number? 1 =)
    (insert 'integer? (check-numbers? 'integer? integer?) 1 =)
    (insert 'boolean? boolean? 1 =)
    (insert 'string? string? 1 =)
    (insert 'symbol? symbol? 1 =)
    (insert 'list? list? 1 =)
    (insert 'null? null? 1 =) ; pas dans carte de reference

    (insert 'positive? (check-numbers? 'positive? positive?) 1 =)
    (insert 'negative? (check-numbers? 'negative? negative?) 1 =)
    (insert 'odd?      (check-integers? 'odd?     odd?) 1 =)
    (insert 'even?     (check-integers? 'even?    even?) 1 =)
    (insert '=         (check-numbers? '=         =) 2 =)
    (insert '+         (check-numbers? '+         +) 0 <=)
    (insert '-         (check-numbers? '-         -) 1 or-one-more?)
    (insert '*         (check-numbers? '*         *) 0 <=)
    (insert '<         (check-numbers? '<         <) 2 =)
    (insert '>         (check-numbers? '>         >) 2 =)
    (insert '<=        (check-numbers? '<=        <=) 2 =)
    (insert '>=        (check-numbers? '>=        >=) 2 =)
    (insert '/         (check-division-numbers? '/         /)2 =)
    (insert 'quotient  (check-division-numbers? 'quotient  quotient) 2 =)
    (insert 'remainder (check-division-numbers? 'remainder remainder) 2 =)
    (insert 'modulo    (check-division-numbers? 'modulo    modulo) 2 =)
    (insert 'random    (check-positive-integers? 'random   random) 1 =)

    (insert 'erreur 
        (lambda args
          (li101-error 1124 args) ) 
        1 <= )
    ;(insert 'erreur? erreur?)
    (insert 'li101-error? 
        (lambda (thunk)
          (with-handler handler2string
                        (begin (thunk)
                               #f ) ) )
        1 = )
    (insert 'string? string? 1 =)
    (global-hashtable-put! h 'li101-print write); DEBUG
    ;; Don't profile li101-, li101-try
    (global-hashtable-put! h 'li101-try li101-try)
    ; This was hack(1) for the old version of (verifier...) now useless!
    ;(global-hashtable-put! h 'li101- 
    ;  ;; exp est le code du test effectué.
    ;  ;; i: nat + (nat string)
    ;  (lambda (i exp) 
    ;    (cond ((integer? i) (- i))
    ;          ;((pair? i) (li101-error (- (car i)) (cdr i)))
    ;          ((pair? i) (li101-error (- (car i))
    ;                                  (->string* "Test malheureux:" exp 
    ;                                             "Erreur:" (cdr i)) ) )
    ;          (else (li101-error "erreur interne! 1199")) ) ) )
    ;; Constants
    (global-hashtable-put! h 'pi        3.141592653589793)
    (global-hashtable-put! h '*epsilon* *epsilon*)
    h ) )
;;; NOTA(1): hack1 est pour distinguer le code de l'etudiant du code
;;; engendre par les tests qu'il ne faut pas prendre en compte pour
;;; comparer les usages.

;;; }}}
;;; {{{ Arbres
(define (ab-vide)
  (vector '**EmptyBinTree**) )
(define (ab? ab)
  (and (vector? ab)
       (let ((vl (vector-length ab)))
         (or (= 1 vl) (= 4 vl)) )
       (memq (vector-ref ab 0)
             '(**BinTree** **EmptyBinTree**) ) ) )
(define (ab-noeud tag left right)
  (if (ab? left)
      (if (ab? right)
          (vector '**BinTree** tag left right)
          (li101-error 1150 'ab-noeud "Fils droit non arbre binaire!") )
      (li101-error 1151 'ab-noeud "Fils gauche non arbre binaire!") ) )
(define (ab-noeud? ab)
  (if (ab? ab)
      (= 4 (vector-length ab))
      (li101-error 1152 'ab-noeud? "Pas un arbre binaire" ab) ) )
(define (ab-etiquette ab)
  (and (ab-noeud? ab)
       (vector-ref ab 1) ) )
(define (ab-gauche ab)
  (and (ab-noeud? ab)
       (vector-ref ab 2) ) )
(define (ab-droit ab)
  (and (ab-noeud? ab)
       (vector-ref ab 3) ) )
(define (ab-affiche ab)
  (if (ab? ab)
      (begin (pp ab) #t) ; pour ne pas rendre #unspecified
      (li101-error 1153 'ab-affiche "Pas un arbre binaire" ab) ) )

;;; bibliotheque arbres generaux
(define (ag-noeud tag forest)
  (if (andmap ag? forest)
      (vector '**GeneralTree** tag forest)
      (li101-error 1154 'ag-noeud 
                   "Un terme de la foret n'est pas un arbre général") ) )
(define (ag? ag)
  (and (vector? ag)
       (= 3 (vector-length ag))
       (eq? (vector-ref ag 0) '**GeneralTree**) ) )
(define (ag-etiquette ag)
  (if (ag? ag)
      (vector-ref ag 1)
      (li101-error 1155 'ag-etiquette "Pas un arbre général" ag) ) )
(define (ag-foret ag)
  (if (ag? ag)
      (vector-ref ag 2)
      (li101-error 1156 'ag-etiquette "Pas un arbre général" ag) ) )
(define (ag-affiche ag)
  (if (ag? ag)
      (begin (pp ag) #t) ; pour ne pas rendre #unspecified
      (li101-error 1157 'ag-affiche "Pas un arbre général" ag) ) )

;;; }}}
;;; {{{ Bibliotheque graphique
(define (check-coordinates coords)
  (define (is-coord? z)
    (or (and (<= -1 z) (<= z 1))
        (li101-error "Pas une coordonnée" z) ) )
  (andmap is-coord? coords) )
(define (check-taille wh)
  (if (not (and (number? wh)
                (<= 0 wh)
                (<= wh 100) ))
      (li101-error "Pas une taille correcte" wh) ) )
(define (image? o)
  (and (vector? o)
       (eq? '**Image** (vector-ref o 0)) ) )
(define (check-image thing)
  (if (image? thing)
      thing
      (li101-error "Pas une image" thing) ) )
(define (check-images images)
  (if (pair? images)
      (if (image? (car images))
          (check-images (cdr images))
          (li101-error "Pas une image" (car images)) )
      #t ) )

(define (make-image w h code)
  (vector '**Image** w h code) )
(define (image-largeur img)
  (check-image img)
  (vector-ref img 1) )
(define (image-hauteur img)
  (check-images img)
  (vector-ref img 2) )

(define (image-vide)
  (make-image 100 100 `(image-vide)) )
(define (check-same-sizes images)
  (let ((hauteur (image-largeur (car images)))
        (largeur (image-hauteur (car images))) )
    (let check ((images (cdr images)))
      (if (pair? images)
          (if (and (= hauteur (image-hauteur (car images)))
                   (= largeur (image-largeur (car images))) )
              (check (cdr images))
              #f ) 
          (list largeur hauteur) ) ) ) )
(define (overlay . images)
  (check-images images)
  (if (pair? images)
      (let ((sizes (check-same-sizes images)))
        (if sizes
            (make-image (car sizes)  ; largeur
                        (cadr sizes) ; hauteur
                        `(overlay . ,images) )
            (li101-error 'overlay "Tailles d'images différentes") ) )
      (li101-error 'overlay "Aucune image") ) )
(define (superposition . images) ; aka overlay  meme img*.w, img*.h
  (check-images images)
  (if (pair? images)
      (let ((sizes (check-same-sizes images)))
        (if sizes
            (make-image (car sizes)  ; largeur
                        (cadr sizes) ; hauteur
                        `(overlay . ,images) )
            (li101-error 'superposition "Tailles d'images différentes") ) )
      (li101-error 'superposition "Aucune image") ) )
(define (ligne x0 y0 x1 y1)
  (check-coordinates (list x0 y0 x1 y1))
  (make-image 100 100 `(ligne ,x0 ,y0 ,x1 ,y1)) )
(define (triangle x0 y0 x1 y1 x2 y2)
  (check-coordinates (list x0 y0 x1 y1 x2 y2))
  (make-image 100 100 `(triangle ,x0 ,y0 ,x1 ,y1 ,x2 ,y2)) )
(define (contour-ellipse x0 y0 x1 y1)
  (check-coordinates (list x0 y0 x1 y1))
  (make-image 100 100 `(contour-ellipse ,x0 ,y0 ,x1 ,y1)) )
(define (ellipse-contour x0 y0 x1 y1)
  (check-coordinates (list x0 y0 x1 y1))
  (make-image 100 100 `(contour-ellipse ,x0 ,y0 ,x1 ,y1)) )
(define (ellipse x0 y0 x1 y1)
  (check-coordinates (list x0 y0 x1 y1))
  (make-image 100 100 `(ellipse ,x0 ,y0 ,x1 ,y1)) )
(define (quart-de-tour img) ; +90 sens horaire
  (check-image img)
  (make-image (image-hauteur img)
              (image-largeur img)
              `(turn ,img) ) )
(define (empilement img1 img2) ; img1.w = img2.w
  (check-image img1)
  (check-image img2)
  (let ((largeur1 (image-largeur img1))
        (largeur2 (image-largeur img2)) )
    (if (= largeur1 largeur2)
        (make-image largeur1
                    (+ (image-hauteur img1) (image-hauteur img2))
                    `(stack ,img1 ,img2) )
        (li101-error 'empilement "Largeurs différentes") ) ) )
(define (symetrique img) ; /y
  (check-image img)
  (make-image (image-largeur img)
              (image-hauteur img)
              `(mirror ,img) ) )
;(define (homothetique img w h)
;  (check-image img)
;  (make-image w h `(resize ,img ,w ,h ,x ,y)) )
;;; Nouvelle image de taille reduite a w,h positionnee avec son coin
;;; inferieur gauche en x,y:
(define (vignette img w h x y)
  (check-image img)
  (check-coordinates (list x y))
  (check-taille w)
  (check-taille h)
  (make-image (image-largeur img)
              (image-hauteur img)
              `(thumbnail ,img ,w ,h ,x ,y)) )
;;; solarise ?

;;; Compile une image en le contenu d'un fichier .eps
(define (image2postscript image)
  (define (coordinate i coords)
    (if (pair? coords)
        (if (= i 0)
            (ceiling (* 100 (car coords)))
            (coordinate (- i 1) (cdr coords)) )
        (li101-error "Internal error" 'coordinate i coords) ) )
  (define (compose . things)
    (define (comp things)
      (if (pair? things)
          (string-append (->string (car things)) 
                         " " (comp (cdr things)) )
          "\n" ) )
    (comp things) )
  (define (translate2ps image)
    (check-images (list image))
    (let ((command (vector-ref image 3)))
      (case (car command)
        ((image-vide) " ")
        ((ligne)
         (let ((x0 (coordinate 0 (cdr command)))
               (y0 (coordinate 1 (cdr command)))
               (x1 (coordinate 2 (cdr command)))
               (y1 (coordinate 3 (cdr command))) )
           (compose "gs"
                    0 0 "tr"
                    "n" x0 y0 "m" x1 y1 "l"
                    "cp" "gs" "col0" "s" "gr"
                    "gr" ) ) )
        ((triangle)
         (let ((x0 (coordinate 0 (cdr command)))
               (y0 (coordinate 1 (cdr command)))
               (x1 (coordinate 2 (cdr command)))
               (y1 (coordinate 3 (cdr command)))
               (x2 (coordinate 4 (cdr command)))
               (y2 (coordinate 5 (cdr command))) )
           (compose "gs" 
                    0 0 "tr"
                    "n" x0 y0 "m" x1 y1 "l" x2 y2 "l"
                    "cp" "gs" "col0" "fill" "gr"
                    "gr" ) ) )
        ((ellipse)
         (let ((x0 (coordinate 0 (cdr command)))
               (y0 (coordinate 1 (cdr command)))
               (x1 (coordinate 2 (cdr command)))
               (y1 (coordinate 3 (cdr command))) )
           (compose "gs" 
                    0 0 "tr"
                    "n" (/ (+ x0 x1) 2) (/ (+ y0 y1) 2) ; centre
                    (/ (- x1 x0) 2) (/ (- y1 y0) 2)     ; rayons
                    0 360 "DrawEllipse"
                    "gs" "col0" "fill" "gr"
                    "gr" ) ) )
        ((contour-ellipse)
         (let ((x0 (coordinate 0 (cdr command)))
               (y0 (coordinate 1 (cdr command)))
               (x1 (coordinate 2 (cdr command)))
               (y1 (coordinate 3 (cdr command))) )
           (compose "gs" 
                    0 0 "tr"
                    "n" (/ (+ x0 x1) 2) (/ (+ y0 y1) 2) ; centre
                    (/ (- x1 x0) 2) (/ (- y1 y0) 2)     ; rayons
                    0 360 "DrawEllipse"
                    "gs" "col0" "stroke" "gr"
                    "gr" ) ) )
        ((overlay) ; superposition
         (let loop ((images (cdr command)))
           (if (pair? images)
               (compose "gs"
                        (translate2ps (car images))
                        "gr"
                        (loop (cdr images)) )
               "" ) ) )
        ((mirror) ; symetrique
         (let* ((image (cadr command))
                (h     (image-hauteur image))
                (w     (image-largeur image)) )
           (compose "gsave" 
                    0 0 "translate"
                    "-1 1 scale"
                    (translate2ps image)
                    "grestore" ) ) )
        ((turn) ; quart-de-tour
         (let* ((image (cadr command)))
           (compose "gs"
                    0 0 "tr"
                    "-90 rot"
                    (translate2ps image)
                    "gr" ) ) )
        ((thumbnail) ; vignette
         (let* ((image (cadr command))
                (oldw  (image-largeur image))
                (oldh  (image-hauteur image))
                (neww  (caddr command))
                (newh  (cadddr command))
                (x  (coordinate 0 (cddddr command)))
                (y  (coordinate 1 (cddddr command))) )
           (compose "gs"  
                    x y "tr"
                    neww newh "tr"
                    (/ neww oldw) (/ newh oldh) "sc"
                    (translate2ps image)
                    "gr" ) ) )
        (else "???") ) ) )
  (let* ((h (image-hauteur image))
         (w (image-largeur image)) )
    (format "%!PS-Adobe-3.0 EPSF-3.0
%%Creator: fig2dev Version 3.2 Patchlevel 5d
%%BoundingBox: -~a -~a ~a ~a
%Magnification: 1.0000
%%EndComments
%%BeginProlog
/$F2psDict 200 dict def
$F2psDict begin
$F2psDict /mtrx matrix put
/col-1 {0 setgray} bind def
/col0 {0.000 0.000 0.000 srgb} bind def
end

/cp {closepath} bind def
/ef {eofill} bind def
/gr {grestore} bind def
/gs {gsave} bind def
/sa {save} bind def
/rs {restore} bind def
/l {lineto} bind def
/m {moveto} bind def
/rm {rmoveto} bind def
/n {newpath} bind def
/s {stroke} bind def
/sh {show} bind def
/slc {setlinecap} bind def
/slj {setlinejoin} bind def
/slw {setlinewidth} bind def
/srgb {setrgbcolor} bind def
/rot {rotate} bind def
/sc {scale} bind def
/sd {setdash} bind def
/ff {findfont} bind def
/sf {setfont} bind def
/scf {scalefont} bind def
/sw {stringwidth} bind def
/tr {translate} bind def
/tnt {dup dup currentrgbcolor
  4 -2 roll dup 1 exch sub 3 -1 roll mul add
  4 -2 roll dup 1 exch sub 3 -1 roll mul add
  4 -2 roll dup 1 exch sub 3 -1 roll mul add srgb}
  bind def
/shd {dup dup currentrgbcolor 4 -2 roll mul 4 -2 roll mul
  4 -2 roll mul srgb} bind def 
/DrawEllipse {
        /endangle exch def
        /startangle exch def
        /yrad exch def
        /xrad exch def
        /y exch def
        /x exch def
        /savematrix mtrx currentmatrix def
        x y tr xrad yrad sc 0 0 1 startangle endangle arc
        closepath
        savematrix setmatrix
        } def

/$F2psBegin {$F2psDict begin /$F2psEnteredState save def} def
/$F2psEnd {$F2psEnteredState restore end} def

/pageheader {
 save
 /W ~a def
 /H ~a def
% newpath -100 -100 moveto -100 +100 lineto +100 +100 lineto +100 -100 lineto
  newpath 0 W sub 0 H sub moveto
          0 W sub H lineto
          W H lineto
          W 0 H sub lineto
 closepath clip newpath
 $F2psBegin
 10 setmiterlimit
 0 slj 0 slc
} bind def

/pagefooter {
 $F2psEnd
 restore
} bind def
%%EndProlog

% marche pas pour inverser les couleurs:
%[{1 exch sub} /exec load currenttransfer /exec load ] cvx settransfer 

pageheader
% here starts figure
1 setlinewidth
~a
% here ends figure
pagefooter
showpage
%%Trailer
%EOF

" w h w h           ; for BoundingBox
  w h               ; for clippath
  (translate2ps image) ) ) )

;;; }}}
;;; {{{ Environment
;;; FUTURE, dans toutes ces primitives, verifier les hypotheses!!!!!!!!!!

(define (andmap pred L)
  (if (pair? L)
      (and (pred (car L))
           (andmap pred (cdr L)) )
      #t ) )

(define (build-primitive-environment)
  (let ((h (build-primitive-environment-arithmetic)))
    (define (insert name function arity comparator)
      (let ((f (check-arity? name function arity comparator)))
        (global-hashtable-put! h name f) ) )
    (insert 'max (check-numbers? 'max max) 1 <=)
    (insert 'min (check-numbers? 'min min) 1 <=)
    (insert 'abs (check-numbers? 'abs abs) 1 =)
    (insert 'sqrt (check-non-negative-numbers? 'sqrt sqrt) 1 =)
    (insert 'pair? pair? 1 =)
    ;; FIXME ces deux definitions ne fonctionnent pas:
    ;(insert 'car (check-pair? 'car car) 1 =)
    ;(insert 'cdr (check-pair? 'cdr cdr) 1 =)
    (insert 'car car 1 =)
    (insert 'cdr cdr 1 =)

    (insert 'caar caar 1 =)
    (insert 'cadr cadr 1 =)
    (insert 'cdar cdar 1 =)
    (insert 'cddr cddr 1 =)
    (insert 'caddr caddr 1 =)
    (insert 'caadr caadr 1 =)
    (insert 'caaar caaar 1 =)
    (insert 'cdaar cdaar 1 =)
    (insert 'cddar cddar 1 =)
    (insert 'cadar cadar 1 =)
    (insert 'cdadr cdadr 1 =)
    (insert 'cdddr cdddr 1 =)

    (insert 'string>? string>? 2 =)

    ;; Pour JP Roy:
    (insert 'first  car 1 =)
    (insert 'rest   cdr 1 =)
    (insert 'second cadr 1 =)
    (insert 'third  caddr 1 =)
    (insert 'fourth cadddr 1 =)
    (insert 'empty? null? 1 =)

    (insert 'cons cons 2 =)
    ;(insert 'list list 0 <=)
    (insert 'list (let ()
                    (define (copy L)
                      (if (pair? L)
                          (begin
                            (incr-profiling 'cons)
                            (cons (car L) (copy (cdr L))) )
                          L ) )
                    (lambda L (copy L)) )
            0 <= )
    ;(insert 'append append 2 =)
    (insert 'append (let ((native-append append))
                      (define (cat . LL)
                        (define (appendN LL)
                          (if (pair? LL)
                              (if (pair? (cdr LL))
                                  (append2 (car LL) (appendN (cdr LL)))
                                  (car LL) )
                              '() ) )
                        (define (append2 L1 L2)
                          (incr-profiling 'cons (length L1))
                          (native-append L1 L2) )
                        (appendN LL) )
                      cat )
            2 <= )
    (insert 'reverse reverse 1 =)
    (insert 'length length 1 =)
    (insert 'member member 2 =)
    (insert 'assoc assoc 2 =)
    (insert 'symbol? symbol? 1 =)
    (insert 'string-length string-length 1 =)
    (insert 'substring substring 2 =)
    (insert 'string-append string-append 2 <=)
    (insert 'vector? vector? 1 =)
    (insert 'vector vector 0 <=)
    (insert 'vector-ref vector-ref 2 =)
    (insert 'vector-length vector-length 1 =)
    (insert 'vector->list vector->list 1 =)
    (insert 'list->vector list->vector 0 <=)
    (insert 'map map 2 =)
    (insert 'display display 1 =)
    (insert 'newline newline 0 =)
    (insert 'filter filter 2 =); changement de nom
    (insert 'filtre filter 2 =); changement de nom
    (insert 'reduce 
            (let () 
              (define (reduce op end L)
                (if (pair? L)
                    (op (car L) (reduce op end (cdr L)))
                    end ) )
              reduce )
            3 =)
    ;; Weird: bigloo (reduce + 1 (list 2 3)) => 5 et pas 6 ?
    ;; Weird: bigloo (reduce cons '() '(a b c)) => (c b . a)

    ;; arbres
    (insert 'ab-vide ab-vide 0 =)
    (insert 'ab-noeud ab-noeud 3 =)
    (insert 'ab-noeud? ab-noeud? 1 =)
    (insert 'ab-etiquette ab-etiquette 1 =)
    (insert 'ab-gauche ab-gauche 1 =)
    (insert 'ab-droit ab-droit 1 =)
    (insert 'ab-affiche ab-affiche 1 =)
    (insert 'ag? ag? 1 =)
    (insert 'ag-noeud ag-noeud 2 =)
    (insert 'ag-etiquette ag-etiquette 1 =)
    (insert 'ag-foret ag-foret 1 =)
    (insert 'ag-affiche ag-affiche 1 =)

    ;; graphique
    (insert 'empty-image     image-vide 0 =)
    (insert 'overlay         overlay 2 <=)
    (insert 'draw-line       ligne 4 =)
    (insert 'fill-triangle   triangle 6 =)
    (insert 'draw-ellipse    contour-ellipse 4 =)
    (insert 'fill-ellipse    ellipse 4 =)
    ;; mes graphiques
    (insert 'image?          image? 1 =)
    (insert 'image-vide      image-vide 0 =)
    (insert 'superposition   superposition 2 <=)
    (insert 'ligne           ligne 4 =)
    (insert 'triangle        triangle 6 =)
    (insert 'ellipse         ellipse 4 =)
    (insert 'contour-ellipse contour-ellipse 4 =)
    (insert 'ellipse-contour ellipse-contour 4 =)
    ;; solarise
    (insert 'quart-de-tour   quart-de-tour 1 =)
    ;; empilement
    (insert 'symetrique      symetrique 1 =)
    (insert 'vignette        vignette 5 =)
    (insert 'image-hauteur   image-hauteur 1 =)
    (insert 'image-largeur   image-largeur 1 =)
    h ) )

(define *the-environment* (build-primitive-environment))

(define (environment-insert! name function . env)
  (let ((env (if (pair? env) (car env) *the-environment*)))
    (global-hashtable-put! env name function)
    env ) )
(define (environment-remove! name . env)
  (let ((env (if (pair? env) (car env) *the-environment*)))
    (hashtable-remove! env name)
    env ) )
(define (environment-get name . env)
  (let ((env (if (pair? env) (car env) *the-environment*)))
    (let ((value (hashtable-get env name)))
      (if value
          (car value)
          (li101-error "No such variable" name) ) ) ) )
(define (environment-duplicate . env)
  (let ((env (if (pair? env) (car env) *the-environment*)))
    (let ((newht (make-hashtable)))
      (map (lambda (key) 
             (hashtable-put! newht key (hashtable-get env key)) )
           (hashtable-key-list env) )
      newht ) ) )
(define (environment-set! env)
  (set! *the-environment* env)
  env )

;;; }}}
;;; {{{ Coverage tests

;;; htable = SExpression -> Alist[exit#, count]
;;; exit# est l'index de la sortie (0 et 1 pour une alternative)
;;; count est le nombre de fois que ce chemin a ete emprunté.
(define husage (create-hashtable :eqtest equal?))
(define (reset-usage!)
  (let ((result husage))
    ;; Return the old hashtable
    (set! husage (create-hashtable :eqtest equal?))
    result ) )
(define (record-usage exp index)
  (debug `(record-usage ,exp ,index));DEBUG
  (let ((alist (or (hashtable-get husage exp) (list))))
    (let ((i+c (or (assoc index alist) (cons index 0))))
      (if (= 0 (cdr i+c))
          (set! alist (cons i+c alist)) ) 
      (set-cdr! i+c (+ 1 (cdr i+c)))
      (hashtable-put! husage exp alist) ) ) )
(define (get-usage husage exp . rest)
  (let search ((keys (hashtable-key-list husage)))
    (if (pair? keys)
        (if (equal? exp (car keys))
            (let ((alist (or (hashtable-get husage (car keys)) (list))))
              (if (pair? rest)
                  (let ((index (car rest)))
                    (let ((i+c (or (assoc index alist) (cons index 0))))
                      (cdr i+c) ) )
                  alist ) )
            (search (cdr keys)) )
        0 ) ) )

;;; On veut que les tests de l'etudiant eprouvent le code de
;;; l'etudiant au moins autant que les tests de l'enseignant.
(define (compare-usage susage tusage fragments)
  ;; [st]usage are hashtables as returned by reset-usage!
  ;; fragments is the list of student's expressions
  (let ((exps (hashtable-key-list tusage))
        (tcover 0)
        (scover 0)
        (hints (list)) )
    (define (inform e) 
      (set! hints (cons e hints)) )
    (for-each 
     (lambda (exp)
       (if (memq exp fragments)
           (let ((talist (hashtable-get tusage exp)))
             (inform `(=== Usage of expression ,exp))
             (debug `(talist ,talist)); DEBUG
             (for-each
              (lambda (ti+c)
                (let ((tindex (car ti+c))
                      (tcount (cdr ti+c)) )
                  (debug `(ti+c ,tindex ,tcount));DEBUG
                  (inform `(===== exitSubcase ,tindex teacherUsage ,tcount))
                  (if (> tcount 0)
                      (let ((scount (get-usage susage exp tindex)))
                        (set! tcover (+ 1 tcover))
                        (debug `(scount ,scount));DEBUG
                        (inform `(======= studentUsage ,scount))
                        (if (> scount 0)
                            (set! scover (+ 1 scover)) ) ) ) ) )
              (sort (lambda (x y)
                      (<= (car x) (car y)) )
                    talist ) ) ) ) )
     exps )
    (list scover tcover (reverse hints)) ) )

;;; Quand on sort d'une alternative par then ou else:
(define (record-then e)
  (record-usage e 0) )
(define (record-else e)
  (record-usage e 1) )
;;; (record-usage '(cond ...) i) 
;;;      signifie que l'on sort de (cond ...) par la ieme clause (0-based)
;;; (record-usage '(and ...) i)
;;;      signifie que les i premieres exp sont vraies et que la
;;;      i+1-eme est fausse (si elle existe)  
;;; (record-usage '(or ...) i)
;;;      signifie que les i premieres exp sont fausses et que la
;;;      i+1-eme est vraie (si elle existe)
;;; }}}

;;; These side-effects may be read with FW4EX_SHOW_BIGLOO_OUTPUT=true
;;; in the fw4ex.xml configuration file.
(define *debug* #f)
(define (debug . args)
  (cond (*debug* (newline)
                 (for-each display args) )
        (else #f) ) )

;;; {{{ Evaluation
(define *eval-counter* 0)
(define (incr-and-get-eval-counter)
  (set! *eval-counter* (+ *eval-counter* 1))
  *eval-counter* )

(define *abort-on-error* #t)
(define *allow-redefinitions* #f)

;;; Tag used to identify values of expressions that are neither a
;;; definition nor a verification:
(define adjoined-value-tag (cons 'adjoined-value 'tag))
;;; Tag used to store complexity measures
(define adjoined-complexity-tag (cons 'adjoined-complexity 'tag))
;;; Tag used to record incrementally the number of successful tests:
(define adjoined-successful-tests-tag (cons 'adjoined-successful-tests 'tag))
;;; Tag to record the complexity measurer (a niladic function):
(define adjoined-measurer-tag (cons 'adjoined-measurer 'tag))
  
;;; li1010-program-eval: List(Expression) * resultat * envGlobal -> resultat
;;; avec resultat := List(Record)
;;;   et Record = FunctionRecord | TaggedRecord
;;;     et FunctionRecord = Nuple(nom fonction checker #tests ...))
;;;       et fonction = (lambda (...) ...) [c'est la valeur de nom pour li101] 
;;;       et checker = (lambda (fonction...) tests) -> Num + (Num excString)
;;;       et #tests nombre de tests ou 0 s'il n'y en a pas
;;;       si #tests>0 alors suivent l'expression (verifier...) et la liste fns
;;;       des fonctions a capturer pour les tests.
;;;     et TaggedRecord = Nuple(tag value ...)
;;;       See adjoined-*-tag for tag values.
;;; ATTENTION, li101-program-eval n'execute pas les tests!
(define (li101-program-eval e* v p)
  (define (eval-define e)
    (if (pair? (cadr e))
        (eval-define-function e)
        (eval-define-constant e) ) )
  (define (eval-define-constant e)
    (let ((name  (cadr e))
          (value (li101-eval (caddr e) (r0))) )
      (global-update! name value p)
      value ) )
  (define (eval-define-function e)
    (let ((name  (car (cadr e)))
          (value (li101-eval `(lambda ,(cdr (cadr e))
                                . ,(cddr e) )
                             (r0) )) )
      (global-update! name value p)
      value ) )
  (define (definition-name def)
    (if (pair? (cadr def))
        (car (cadr def))
        (cadr def) ) )
  (define (definition-expression def)
    (if (pair? (cadr def))
        `(lambda ,(cdr (cadr def)) . ,(cddr def))
        (caddr def) ) )

  ;; eval-verify as a li101 macro:
  (define (XXXeval-verify e)
    (define (convert e* i)
      (if (pair? e*)
          (let ((exp1 (car e*))
                (kind (cadr e*))
                (exp2 (caddr e*))
                (rest (cdddr e*)) )
            (case kind
              ((=> -> ~>)
               ;; expression -> nat + (nat . "erreur")
               ;; nat est positif si tous les tests sont reussis
               ;; si nat est negatif, c'est le numero du 1er test raté.
               (if (memq exp2 '(erreur ERREUR error ERROR))
                   `(if (li101-error? (lambda () ,exp1))
                        ,(convert rest (+ i 1))
                        (li101- ,i ',exp1) )
                   ; `(if (equal? ,exp1 ',exp2)
                   ;      ,(convert rest (+ i 1))
                   ;      (li101- ,i ',exp1) )
                   `(let ((li101-result (li101-try (lambda () ,exp1))))
                      ;(li101-print li101-result);DEBUG
                      (if (if (pair? (cdr li101-result)) #f #t)
                          (if (equal? (car li101-result) ',exp2)
                              ,(convert rest (+ i 1))
                              (li101- ,i ',exp1) )
                          (li101- (cons ,i li101-result) ',exp1) ) ) ) )
              ;; Inutile maintenant:
              ((~>)
               `(if (let ((result ,exp1)
                          (expected ',exp2) )
                      (<= (abs (- result expected)) *epsilon*) )
                    ,(convert rest (+ i 1))
                    (li101- ,i ',exp1) ) ) ) )
; interdit aux etudiants donc aux enseignants pour leurs tests:
;              ((==) 
;               `(if (equal? ,exp1 ,exp2)
;                    ,(convert rest (+ i 1))
;                    (li101- ,i) ) ) ) )
          `,i ) )
    (let* ((tests (quotient (length (cddr e)) 3))
           ;; On code que le n-ieme test est erroné par -n
           ;; Si les n tests sont satisfaits, on rend n+1
           ;; Si aucun test n'est satisfait, on rend -1
           (check-expression (convert (cddr e) 1))
           (fns (if (pair? (cadr e)) (cadr e) (list (cadr e))))
           (checker-code `(lambda ,fns ,check-expression)) )
      (let ((checker (li101-eval checker-code (r0))))
        ;; Attention, les tests ne sont pas effectues, seule la
        ;; fonction de verification est engendree.
        (list checker tests check-expression fns) ) ) )

  ;; Exemple de generation pour 
  ;; (define (foo x) (sqrt x))
  ;; (verifier foo
  ;;    (foo 4) ~> 2
  ;;    (foo 16.1) ~> 4.01 )
  ;; produit:
  ;; (let ((li101-result (li101-try (lambda () (foo 4)))))
  ;;   (if (pair? result)
  ;;       (if (equal? (car li101-result) '2)
  ;;           (let ((li101-result (li101-try (lambda () (foo 16.1)))))
  ;;             (if (pair? result)
  ;;                 (if (equal? (car li101-result) '4.01)
  ;;                     3
  ;;                     (li101- 2 '(foo 16.1)))
  ;;                 (li101- (cons 2 li101-result) '(foo 16.1))))
  ;;           (li101- 1 '(foo 4)))
  ;;       (li101- (cons 1 li101-result) '(foo 4))))

  ;; a new definition of verify as a keyword
  (define (eval-verify e)
    ;; (verify f|(f ...) exp => value ...)
    ;; as checked in the syntax, verify is a top-level form.
    (define (eval-clause exp1 kind exp2 count r)
      ;; kind is always in *test-symbols*
      (if (memq exp2 '(erreur ERREUR error ERROR))
          (eval-erroneous-clause exp1 count r)
          (eval-normal-clause exp1 exp2 count r) ) )
    ;; eval-erroneous-clause: ... -> Num
    (define (eval-erroneous-clause exp count r)
      (debug `(** eval-erroneous-clause ,exp));DEBUG
      (let ((result (li101-try (lambda () (li101-eval exp r)))))
        ;; result = (value) | (msg exc)
        (if (pair? (cdr result))
            ;; an error was expected so the test is successful:
            count
            ;; no error occurred, the test failed:
            (- count) ) ) )
    ;; eval-normal-clause: ... -> Num + (-Num string)
    (define (eval-normal-clause exp1 exp2 count r)
      (debug `(** eval-normal-clause ,exp1));DEBUG
      (let ((result (li101-try (lambda () (li101-eval exp1 r)))))
        ;; result = (value) | (msg exc)
        (if (pair? (cdr result))
            ;; no error was expected, the test failed but report the error:
            (let ((s (string-append (car result) " " (cadr result))))
              (list (- count) s) )
            (if (quasi-equal? (car result) exp2)
                count
                ;; not the expected result: the test failed:
                (- count) ) ) ) )
    ;; eval-clauses: ... -> Num OR throw (Num + string)
    (define (eval-clauses e* count total r)
      (if (pair? e*)
          (let ((exp1 (car e*))
                (kind (cadr e*))
                (exp2 (caddr e*))
                (rest (cdddr e*)) )
            (let ((testresult (eval-clause exp1 kind exp2 count r)))
              (debug (list 'testresult testresult)); DEBUG
              ;; testresult = Num + (Num exceptionString)
              (cond ((and (number? testresult) (> testresult 0))
                     ;; Record the partial result:
                     (adjoin-progress! testresult total) ;; Encore inutile????
                     (eval-clauses rest (+ count 1) total r) )
                    ((and (number? testresult) (< testresult 0))
                     ;; last test failed:
                     testresult )
                    ((and (pair? testresult)
                          (number? (car testresult)) )
                     ;; could not run some badly written test:
                     (li101-throw-error (car testresult) (cadr testresult)) )
                    (else problem13) ) ) )
          count ) )
    (let* ((tests (quotient (length (cddr e)) 3))
           ;; On code que le n-ieme test est erroné par -n
           ;; Si les n tests sont satisfaits, on rend n+1
           ;; Si aucun test n'est satisfait, on rend -1
           ;; si une anomalie survient, on rend (-n exception)
           ;; ou n est le numero du test raté et exception 
           ;; une chaine representant l'exception.
           (fns (if (pair? (cadr e)) (cadr e) (list (cadr e))))
           (checker (eval `(lambda ,fns 
                             ;; HACK: use current pairlis, r0 and eval-clauses
                             (let ((r (,pairlis (,r0) 
                                               ',fns
                                               (list ,@fns) )))
                               (,eval-clauses ',(cddr e) 1 ,tests r) ))) ) )
      ;; Attention, les tests ne sont pas effectues, seule la
      ;; fonction de verification est engendree.
      (list checker tests e fns) ) )

  ;; Estimate complexity
  (define (eval-li101-verify exp)
    ;; (li101-verify (f ...) (fns...) int => exp ...)
    ;; as checked in the syntax, li101-verify is a top-level form.
    (let ((fns       (cadr exp))   ; functions to profile
          (functions (caddr exp))  ; functions to capture and invoke
          (clauses   (cdddr exp)) )
      (define (eval-clause! index exp r)
        (reset-profiling!)
        (li101-eval exp r)
        (let ((counters (map (lambda (name)
                               (list name (get-profiling name)) )
                             fns )))
          (debug `(+++ counters ,index ,counters)); DEBUG
          (cons index counters) ) )
      (define (eval-clauses! clauses r)
        (if (pair? clauses)
            (cons (eval-clause! (car clauses) (caddr clauses) r)
                  (eval-clauses! (cdddr clauses) r) )
            (list) ) )
      (define (measurer)
        (eval `(lambda ,functions
                 (let* ((functions-values (list ,@functions))
                        (r (,r0))
                        (r (,pairlis r ',functions functions-values))
                        (measures (,eval-clauses! ',clauses r)) )
                   ;(pp `(measures ,measures)) ; DEBUG
                   ;; yields Alist(Nuple(index, Alist(funName, nat)))
                   measures ) )))
      (measurer) ) )

  (define (whatever) 42)
  (define (li101-eval e r)
    (let ((counter (incr-and-get-eval-counter)))
      (debug `(li101-eval (,counter) ,e ,r))
      (let ((value (if (pair? e)
                       (case (car e)
                         ((quote)  (eval-quote e r))
                         ((if)     (eval-if e r))
                         ((cond)   (eval-cond e r))
                         ((let)    (eval-let e r))
                         ((let*)   (eval-let* e r))
                         ((begin)  (eval-begin e r))
                         ((lambda) (eval-lambda e r))
                         ((and)    (eval-and e r))
                         ((or)     (eval-or e r))
                         ;((@eval)  (eval-scheme e r))
                         (else     (eval-application e r)) )
                       (cond ((symbol? e) (lookup e r p))
                             (else e) ) ) ))
        (debug `(li101-value (,counter) ,value))
        value ) ) )
  ;; un echappement vers le Scheme sous-jacent. N'a pas encore servi!
  (define (eval-scheme e r)
    ;; r is useless
    (eval `(begin . ,e))
    'done )
  (define (eval-and e r)
    (define (eval-and* e* last i)
      (incr-profiling 'and)
      (if last
          (if (pair? e*)
              (eval-and* (cdr e*) 
                         (li101-eval (car e*) r)
                         (+ i 1) )
              last )
          (begin
            (record-usage e i)
            last ) ) )
    (eval-and* (cdr e) #t 0) )
  (define (eval-or e r)
    (define (eval-or* e* i)
      (incr-profiling 'or)
      (if (pair? e*)
          (let ((value (li101-eval (car e*) r)))
            (if value
              (begin (record-usage e i)
                     value )
              (eval-or* (cdr e*) (+ i 1)) ) )
          (begin 
            (record-usage e i)
            #f ) ) )
    (eval-or* (cdr e) 0) )
  (define (eval-application e r)
    (let* ((f    (li101-eval (car e) r))
           (args (li101-eval* (cdr e) r)) )
      (apply f args) ) )
  (define (li101-eval* e* r)
    (if (pair? e*)
        (let ((value (li101-eval (car e*) r)))
          (cons value (li101-eval* (cdr e*) r)) )
        (list) ) )
  (define (eval-begin e r)
    (if (pair? e)
        (if (pair? (cdr e))
            ;; On ne compte que les vrais (begin...):
            (begin (incr-profiling 'begin)
                   (li101-eval (car e) r)
                   (eval-begin (cdr e) r) )
            (li101-eval (car e) r) )
        ;; Syntax forbids that (if syntax is checked!):
        (whatever) ) )
  (define (eval-quote e r)
    (cadr e) )
  (define (eval-if e r)
    ;; (assume (pair? e))
    (incr-profiling 'if)
    (cond ((li101-eval (cadr e) r)
           (record-then e)
           (li101-eval (caddr e) r) )
          ((pair? (cdddr e))
           (record-else e)
           (li101-eval (cadddr e) r) )
          (else 
           (record-else e)
           (whatever) ) ) )
  (define (eval-cond e r)
    (define (eval-clauses e* r i)
      (incr-profiling 'cond)
      (if (pair? e*)
          (let ((condition (car (car e*)))
                (body      (cdr (car e*))) )
            (if (or (eq? condition 'else)
                    (li101-eval condition r) )
                (begin 
                  (record-usage e i)
                  (eval-begin body r) )
                (eval-clauses (cdr e*) r (+ i 1)) ) )
          (whatever) ) )
    (eval-clauses (cdr e) r 0) )
  (define (eval-let e r)
    (incr-profiling 'let)
    (let ((bindings (cadr e))
          (body     (cddr e)) )
      (let ((names  (map car bindings))
            (values (map (lambda (binding)
                           (li101-eval (cadr binding) r) )
                         bindings )) )
      (eval-body+define body (pairlis r names values)) ) ) )                  
  (define (eval-let* e r)
    (incr-profiling 'let*)
    (let ((bindings (cadr e))
          (body     (cddr e)) )
      (define (bind bindings r)
        (if (pair? bindings)
            (let ((name (car (car bindings)))
                  (exp  (cadr (car bindings))) )
              (bind (cdr bindings)
                    (extend r name (li101-eval exp r)) ) )
            r ) )
      (eval-body+define body (bind bindings r)) ) )
  (define (eval-body+define e* r)
    (define (extract-definitions e* k)
      ;; (assume (pair? e*))
      (if (is-define? (car e*))
          (extract-definitions 
           (cdr e*)
           (lambda (defs body)
             (k (cons (car e*) defs) body) ) )
          (k (list) e*) ) )
    (extract-definitions 
     e*
     (lambda (definitions body)
       (let ((r2 (extend* r (map definition-name definitions))))
         (map (lambda (def)
                (update! r2 
                         (definition-name def)
                         (li101-eval (definition-expression def) r2) ) )
              definitions )
         ;; special index for body entry:
         (record-usage body 99999)
         (eval-begin body r2) ) ) ) )
  (define (pairlis r variables values)
    (if (pair? variables)
        (if (pair? values)
            (cons (cons (car variables) (car values))
                  (pairlis r (cdr variables) (cdr values)) )
            (li101-error 1125 
                         "Il manque l'argument correspondant à la variable" 
                         (car variables) ) )
        (if (pair? values)
            (li101-error 1126 "Argument superflu" (car values))
            r ) ) )
  (define (eval-lambda e r)
    (let ((variables (cadr e))
          (body      (cddr e)) )
      (lambda args
        (let ((r2 (pairlis r variables args)))
          (eval-body+define body r2) ) ) ) )

  (define (r0) (list))
  (define (lookup e r p)
    (if (pair? r)
        (if (eq? e (caar r))
            (cdar r)
            (lookup e (cdr r) p) )
        (global-lookup e p) ) )
  (define (extend r e v)
    (cons (cons e v) r) )
  (define (extend* r e*)
    (if (pair? e*)
        (cons (cons (car e*) 111)
              (extend* r (cdr e*)) )
        r ) )
  (define (update! r e v)
    (if (pair? r)
        (if (eq? e (caar r))
            (set-cdr! (car r) v)
            (update! (cdr r) e v) )
        (li101-error 1127 "Je ne peux mettre à jour la variable" e) ) )

  (define (global-lookup e p)
    (let ((value (hashtable-get p e)))
      (if value
          (car value)
          (li101-error 1128 "Variable indéfinie" e) ) ) )
  (define (global-update! e v p)
    (let ((value (hashtable-get p e)))
      (if value
          ;; There's already a value!
          (if *allow-redefinitions*
              (global-hashtable-put! p e v)
              (li101-error 1132 "Redéfinition interdite!" e) )
          (global-hashtable-put! p e v) ) )
    v )
  (define (global-profile! name p)
    (let ((f (global-lookup name p)))
      (if f
          (let ((newf (make-profiling name f)))
            (global-hashtable-put! p name newf) ) ) ) )

  ;; Chaque definition fait l'objet d'un Tuple (nom fonction verifieur 0)
  ;; Les fonctions de LI101 sont implantées par des fonctions Scheme acceptant
  ;; la meme arité. Le Tuple est initialise avec un verifieur toujours faux.
  ;; Attention, ce tuple sera modifie lorsque le verifieur associe sera
  ;; defini et alors, le nombre de tests ne sera plus 0 et certains champs
  ;; additionnels seront presents.
  (define (adjoin-function name value v)
    (define (checker n)
      (li101-error 1129 "Vérification absente pour" name) )
    (if (assq name v)
        (li101-error 1130 "Redéfinition de" name)
        (cons (list name value checker 0) v) ) )
  ;; Chaque verification modifie les tuples des definitions concernees qui 
  ;; deviendront (nom fonction verifieur #tests codeVerifieur fonctionsTestees)
  (define (adjoin-checker names value+n+code+fns v)
    (define (adjoin name)
      (let ((nuple (assq name v)))
        (if nuple
            (if (not (= 0 (cadddr nuple)))
                (li101-error 1140 "Redéfinition des tests de" name)
                (begin
                  (debug `(adjoin-checker ,v));DEBUG
                  (set-cdr! (cdr nuple) value+n+code+fns)
                  (debug `(adjoin-checker ,v));DEBUG
                  v ) )
            ;; les formes (verifier ...) doivent arriver apres les definitions:
            (li101-error 1131 "Fonction indéfinie" name) ) ) )
    (if (pair? names)
        (for-each adjoin names)
        (adjoin names) )
    v )
  (define (adjoin-measurer fn+form v)
    (adjoin-tagged-value adjoined-measurer-tag fn+form v) )
  ;; Il y a aussi d'autres Tuples specifiques dont le premier terme 
  ;; n'est pas un symbole. See the list of tags (adjoined-*-tag).
  (define (adjoin-tagged-value tag value v)
    (let ((nuple (assq tag v)))
      (if nuple 
          (begin 
            (set-cdr! nuple (cons value (cdr nuple)))
            v )
          (cons (cons tag (list value)) v) ) ) )
  ;; On stocke aussi les valeurs des expressions qui ne sont ni des
  ;; definitions ni des verifications. C'est utile pour recuperer un
  ;; dessin qui doit figurer en derniere expression.
  (define (adjoin-value value v)
    (adjoin-tagged-value adjoined-value-tag value v) )
  ;; Stocker la progression des tests telle qu'enregistree par eval-verify.
  ;; NOTA: ici, on ne sait de quelle configuration de test il s'agit!
  (define (adjoin-progress! count total)
    (set! v (cons (list adjoined-successful-tests-tag count total) 
                  v ))
    v )
  (define (adjoin-profiling! index counters)
    (set! v (cons (cons adjoined-complexity-tag (cons index counters)) 
                  v ))
    v )

  (debug `(+++ entering li101-program-eval ,e*))
  (if (pair? e*)
      (cond ((is-define-function? (car e*))
             (let ((name  (definition-name (car e*)))
                   (value (eval-define-function (car e*))))
               (li101-program-eval (cdr e*)
                                   (adjoin-function name value v)
                                   p ) ) )
            ((is-define-constant? (car e*))
             (let ((value (eval-define-constant (car e*))))
               (li101-program-eval (cdr e*) v p) ) )
            ((is-verify? (car e*))
             (let* ((names (cadr (car e*)))
                    (checker+n+code+fns (eval-verify (car e*))) )
               ;; names peut etre une liste de noms a tester ensemble.
               (li101-program-eval (cdr e*)
                                   (adjoin-checker names checker+n+code+fns v)
                                   p ) ) )
            ((is-li101-verify? (car e*))
             ;; compute complexity:
             (let ((measurer (eval-li101-verify (car e*))))
               (li101-program-eval (cdr e*)
                                   (adjoin-measurer (list measurer (car e*)) v)
                                   p ) ) )
            (else
             (let ((value (li101-eval (car e*) (r0))))
               ;; Stocker la valeur de cette expression:
               (li101-program-eval (cdr e*) 
                                   (adjoin-value value v)
                                   p) ) ) )
      (begin 
        (debug `(+++ exiting li101-program-eval ,v))
        v ) ) )

;;; }}}
;;; {{{ Analyse complexite


;;; }}}
;;; {{{ Evaluation et comparaison de deux fichiers

;;; Enregistrer toutes les expressions du code de l'etudiant. 
;;; Il faudra tester l'appartenance avec memq. C'est necessaire afin
;;; de determiner le taux de couverture des tests de l'etudiant en omettant
;;; tout fragment de code qui viendrait de l'enseignant.
(define (collect-student-code student)
  (let ((result (list)))
    (define (push e)
      (set! result (cons e result)) )
    (define (walk e)
      (if (pair? e)
          (begin (push e)
                 (walk (car e))
                 (walk (cdr e)) ) ) )
    (walk student)
    result ) )               

;;; eval-and-compare: FileName * FileName -> List(Nuple)
;;; with Nuple := (string "message") | (data kind value)
;;;       | (error "ERREUR: message") | (warning "msg")
;;;       | (data student-total-tests nat)
;;;       | (data teacher-total-tests nat)
;;;       | (data student-good-tests nat)
;;;       | (data teacher-good-tests nat)
;;;       | (data compare-usage (nat nat))
;;;       | (data teacher-profiling ...)
;;;       | (data student-profiling ...)
;;; the remaining arguments may first specify the environment 
;;; then the logger to use.
(define (eval-and-compare student-file teacher-file . others)
  ;; By default, use *the-environment*
  (define env (if (pair? others)
                  (let ((env (car others)))
                    (set! others (cdr others))
                    env )
                  *the-environment* ))
  (define student-env 'wait)
  (define teacher-env 'wait)
  ;; Incremental logger. Attention, it receives a reverse log.
  (define logger (if (pair? others)
                     (car others)
                     (lambda (reverselog) reverselog) ))
  (define log (list))
  (reset-profiling!)
  (reset-usage!)
  (set! student-env env)
  (set! teacher-env (environment-duplicate env))
  (call/cc 
   (lambda (exit)
     (define (->single-string msgs)
       (apply string-append 
              (map (lambda (o) 
                     (string-append " " (->string o)) )
                   msgs ) ) )
     (define (abort . msgs)
       (let ((s (->single-string msgs)))
         (debug `(*** error ,s))
         (set! log (cons (list 'error s) log))
         (logger log)
         (exit (reverse log)) ) )
     (define (record kind value)
       (debug `(*** data ,kind ,value))
       (set! log (cons (list 'data kind value) log))
       (logger log) )
     (define (push . msgs)
       (let ((s (->single-string msgs)))
         (debug `(*** string ,s))
         (set! log (cons (list 'string s) log))
         (logger log) ) )
     (define (warn . msgs)
       (let ((s (->single-string msgs)))
         (debug `(*** warning ,s))
         (set! log (cons (list 'warning s) log))
         (logger log) ) )
     (define warnings #f) 
     (define (warn-or-abort . words)
       (set! warnings #t)
       (if *abort-on-error*
           (apply abort (append words (list "je m'arrête là")))
           (apply warn words) ) )

     ;; Chercher dans v* les valeurs des fonctions nommees dans le
     ;; champ fns. Substituer fname a la fonction name. Retourner la
     ;; liste de ces fonctions.
     (define (collect-checked-functions v* v name fname)
       (let ((fns (collect-checked-functions-names v)))
         (debug `(fns ,fns));DEBUG
         (map (lambda (othername)
                (if (eq? name othername)
                    fname
                    (let ((otherv (assq othername v*)))
                      (if (pair? otherv)
                          (cadr otherv)
                          (lambda things
                            (li101-error 1133
                               "Je ne trouve pas votre définition de"
                               "la fonction" othername) ) ) ) ) )
              fns ) ) )
     (define (collect-checked-functions-names v)
       (caddr (cdddr v)) )
     (define (collect-named-functions v* fns)
       (map (lambda (name)
              (let ((v (assq name v*)))
                (if (pair? v)
                    (cadr v)
                    #f ) ) )
            fns ) )

     ;; lire le fichier de l'étudiant:
     (define (read-student-file student-file)
       (push "Je vais lire votre fichier")
       (let ((student (with-handler handler2string
                                    (slurp-file student-file) )))
         (if (string? student)
             (abort "Votre fichier ne peut être lu (" student ")")
             student ) ) )

     ;; le vérifier syntaxiquement:
     (define (check-student-file student)
       (push "J'analyse la syntaxe de votre fichier")
       (let ((student-code (with-handler handler2string
                                         (check-program? student) )))
         (if (string? student-code)
             (abort "Votre fichier est syntaxiquement incorrect ("
                    student-code ")" )
             student ) ) )

     ;; l'évaluer. Si le fichier s'evalue correctement. Attention, les
     ;; tests de l'etudiant ne sont pas encore executes.
     (define (eval-student-file student-code)
       (push "J'évalue votre fichier")
       (let ((student-v (with-handler handler2string
                             (li101-program-eval 
                              student-code
                              (list)
                              student-env ) )))
         (if (string? student-v)
             ;; Dans ce cas stt sera nul.
             (abort "Votre fichier s'évalue erronément (" student-v ")")
             (let* ((funsv (filter (lambda (v)
                                     (symbol? (car v)) )
                                   student-v ))
                    (names (map car funsv))
                    (ntests (apply + (map cadddr funsv))) )
               (cond
                ((= 1 (length names))
                 (push "Vous avez défini la fonction " (car names)) )
                (else
                 (apply push "Vous avez défini les fonctions suivantes "
                        (reverse names) ) ) )
               ;; enregistrer le nombre total de tests ecrits par l'etudiant:
               (record 'student-total-tests ntests) ) )
         student-v ) )
     ;; NOTA: Verifier que l'etudiant a bien ecrit des tests pour les
     ;; fonctions demandees ne peut etre fait ici car on ne sait pas
     ;; encore les fonctions qui sont demandees. Il faut pour cela
     ;; avoir lu le fichier de l'enseignant.

     ;; lire le fichier symétrique de l'enseignant:
     (define (read-teacher-file teacher-file)
       (push "Je lis maintenant le fichier contenant ma solution")
       (let ((teacher (with-handler handler2string
                                    (slurp-file teacher-file) )))
         (if (string? teacher)
             (abort "Le fichier exercice ne peut être lu (" 
                    teacher ")" )
             teacher ) ) )
     ;; vérifier s'il appartient au dialecte li101:
     ;; En fait, l'enseignant a le droit a un langage plus riche: 
     ;; le langage d'entree de li101-program-eval.

     (define (check-teacher-file teacher)
       ;; (push "J'analyse la syntaxe de ma solution")
       ;; (let ((teacher-code (with-handler handler2string
       ;;                                   (check-program? teacher) )))
       ;;   (if (string? teacher-code)
       ;;       (abort "Le fichier exercice est syntaxiquement incorrect ("
       ;;              teacher-code ")" )
       ;;       teacher ) )
       teacher )

     ;; evaluer le code de l'enseignant hors tests enseignant:
     (define (eval-teacher-file teacher-code)
       (push "J'évalue ma solution")
       (let ((teacher-v (with-handler handler2string
                             (li101-program-eval 
                              teacher-code
                              (list)
                             teacher-env ) )))
         ;; NOTA: il faut dupliquer l'environnement initial afin que les
         ;; fonctions de l'enseignant ne viennent pas remplacer celles de
         ;; l'etudiant.
         (if (string? teacher-v)
             (abort "Mon fichier s'évalue incorrectement (" teacher-v ")")
             (let* ((funsv (filter (lambda (v)
                                    (symbol? (car v)) )
                                  teacher-v ))
                    (names  (map car funsv))
                    (ntests (apply + (map cadddr funsv))) )
               (cond 
                ((= 1 (length names))
                 (push "J'ai, quant à moi, défini la fonction "
                        (car names) ) )
                (else
                 (apply push "J'ai, quant à moi, défini les fonctions "
                        (reverse names) ) ) )
               ;; enregistrer le nombre total de tests ecrits par l'enseignant:
               (record 'teacher-total-tests ntests) ) )
         teacher-v ) )

     ;; On passe les fonctions de l'enseignant dans ses propres tests.
     (define (check-teacher-own-tests teacher-v)
       (define (check-one-function tv name tf tchecker n)
         (define (verbalise result errmsg)
           (cond ((and (number? result) (< result 0))
                  (let ((ntests (- (abs result) 1)))
                    (record 'teacher-good-own-tests ntests)
                    (cond 
                     ((= ntests 0)
                      (abort "Ma fonction" name "ne passe aucun de mes tests"
                             (if errmsg errmsg "") "je m'arrête là" ) )
                     ((= ntests 1)
                      (abort "Ma fonction " name "ne passe que le"
                             "premier de mes tests" (if errmsg errmsg "")
                             "je m'arrête là" ) )
                     (else
                      (abort "Ma fonction" name "échoue sur mon test numéro"
                             ntests (if errmsg errmsg "")
                             "je m'arrête là" ) ) ) ) )
                 (errmsg
                  (if (pair? (cdr names))
                      (abort "Mes fonctions" names "ne passent pas"
                             "mes propres" n "tests (" errmsg ")" )
                      (abort "Ma fonction" name "est erronée et"
                             "ne passe pas mes propres" n "tests ("
                             errmsg ")" ) ) ) ) )
         (let* ((names (collect-checked-functions-names tv))
                (fns   (collect-named-functions teacher-v names)) )
            (if (andmap procedure? fns)
                (let* ((result (with-handler handler2pair
                                     (list (apply tchecker fns)) ))
                       ;; result: (n) + (n errmsg)
                       ;(void1 (debug `(check-teacher-own-tests ,result)));DEBUG
                       ;; errmsg: Exception + #f
                       (errmsg (cond ((string? result)
                                      result )
                                     ((and (pair? result)
                                           (pair? (cdr result)) )
                                      (cadr result) )
                                     (else #f) ))
                       ;; result: nat (positif si ok, negatif sinon)
                       (result (cond ((string? result)
                                      -1 )
                                     ((and (pair? result)
                                           (pair? (cdr result)) )
                                      (car result) )
                                     (else (car result)) )) )
                  (verbalise result errmsg)
                  (record 'teacher-good-own-tests (- result 1))
                  (push "Ma fonction" name "passe tous mes" n "tests")
                  result ) ) ) )
       (define (run-one-verification tv)
          (if (symbol? (car tv))
              (let* ((name     (car tv))
                     (tf       (cadr tv))
                     (tchecker (caddr tv))
                     (n        (cadddr tv)) )
                (if (> n 0)
                    (check-one-function tv name tf tchecker n)
                    (warn "Je n'ai pas défini de tests pour ma fonction"
                          name ) ) ) ) )
       (push "Je vais tester mes fonctions avec mes propres tests.")
       (for-each run-one-verification teacher-v)
       ;(push "J'ai fini de tester mes fonction avec mes tests.");DEBUG
       teacher-v )

     ;; Les fonctions de l'enseignant doivent satisfaire les 
     ;; tests de l'etudiant. Les seules fonctions a tester sont celles
     ;; pour lesquelles l'enseignant a ecrit une forme (verifier f ...)
     ;; ou (verifier (f ...) ...)
     ;; Attention: les tests de l'etudiant peuvent verifier des proprietes
     ;; qui ne sont pas dans l'enonce, la fonction de l'enseignant a alors
     ;; peu de chance de les reussir. Aussi rater cette phase ne compte pas.
     (define (check-student-tests student-v teacher-v)
       (define (check-one-function sv name sf schecker n)
         (define (verbalise result errmsg)
           (cond ((and (number? result) (< result 0))
                  (let ((ntests (- (abs result) 1)))
                    (record 'student-good-tests ntests)
                    (cond
                     ;; on ne montre pas errmsg!
                     ((= ntests 0)
                      (warn "Ma fonction" name "ne passe"
                            "pas le premier de vos" n "tests." 
                            "Mais je continue!" ) )
                     ((= ntests 1)
                      (warn "Ma fonction" name "ne passe"
                            "que le premier de vos" n "tests." 
                            "Mais je continue!" ) )
                     (else
                      (warn "Ma fonction" name "échoue à" 
                            "votre test numéro" ntests 
                            "mais je continue!" ) ) ) ) )
                 (errmsg (warn "Ma solution à" name "ne satisfait pas vos"
                               n "tests (" errmsg ")" ))
                 (else
                  (record 'student-good-tests (- result 1))
                  (push "Ma fonction" name "passe tous vos" n "tests")
                  result ) ) )
         (let* ((names (collect-checked-functions-names sv))
                (fns   (collect-named-functions teacher-v names)) )
           (if (andmap procedure? fns)
               (let* ((result (with-handler handler2pair
                                   (list (apply schecker fns)) ))
                      ;; result: (value) + (n errmsg)
                      ;(void1 (debug `(==== check-student-tests 
                      ;                     after ,name ,result))) ;DEBUG
                      ;; errmsg: Exception + #f
                      (errmsg (cond ((string? result)
                                     result )
                                    ((and (pair? result)
                                          (pair? (cdr result)) )
                                     (cadr result) )
                                    (else #f) ))
                      ;; result: nat (positif si ok, negatif sinon)
                      (result (cond ((string? result)
                                     -1 )
                                    ((and (pair? result)
                                          (pair? (cdr result)) )
                                     (car result) )
                                    (else (car result)) )) )
                 (verbalise result errmsg) )
               (warn "Votre test porte sur au moins une fonction "
                     "non demandée par l'énoncé" ) ) ) )
       (define (run-one-verification sv)
         (let* ((name     (car sv))
                (sf       (cadr sv))
                (schecker (caddr sv))
                (n        (cadddr sv)) )
           (debug `(==== check-student-tests ,name)) ;DEBUG
           (if (> n 0)
               (check-one-function sv name sf schecker n) ) ) )
       (push "Je teste à présent que ma solution passe vos tests")
       ;; pour chaque verification de l'etudiant:
       (for-each (lambda (sv)
                   (if (symbol? (car sv))
                       (run-one-verification sv) ) )
                 student-v ) )

     ;; On passe les fonctions de l'etudiant dans ses propres test
     ;; afin de mesurer son usage.
     (define (check-students-own-tests student-v)
       (define (check-one-function sv name sf schecker n)
         (define (verbalise result errmsg counter)
           (cond
            ((and (number? result) (< result 0))
             (let ((ntests (- (abs result) 1)))
               ;; nombre de tests reussis:
               (record 'student-good-own-tests ntests)
               (cond
                ((= ntests 0)
                 (abort "Votre fonction" name
                        "ne passe aucun de vos tests"
                        (if errmsg errmsg "") ) )
                ((= ntests 1)
                 (abort "Votre fonction " name "ne passe que le"
                        "premier de vos tests"
                        (if errmsg errmsg "") ) )
                (else
                 (abort "Votre fonction" name
                        "échoue sur votre test numéro"
                        (+ ntests 1) "je m'arrête là"
                        (if errmsg errmsg "") ) ) ) ) )
            (errmsg
             (abort "Votre fonction" name "est erronée et"
                    "ne passe pas vos propres" n "tests ("
                    errmsg ")" ) )
            ((= counter 0)
             (record 'student-good-own-tests (- result 1))
             (abort "Vos tests n'utilisent pas votre fonction"
                    name ) )
            (else
             (record 'student-good-own-tests (- result 1))
             (push "Votre fonction" name
                   "passe tous vos" n "tests" )
             result ) ) )
         (let* ((counter 0)
                (scountingf (lambda args
                              (set! counter (+ 1 counter))
                              (apply sf args) ))
                (names (collect-checked-functions-names sv))
                (fns   (collect-checked-functions
                        student-v sv name scountingf ))
                ;; result: (value) + (n errmsg)
                (result (with-handler handler2pair
                                      (list (apply schecker fns)) ))
                ;; errmsg: Exception + #f
                (errmsg (cond ((string? result)
                               result )
                              ((pair? (cdr result))
                               (cadr result) )
                              (else #f) ))
                ;; result: nat (positif si ok, negatif sinon)
                (result (cond ((string? result)
                               -1 )
                              ((pair? (cdr result))
                               (car result) )
                              (else (car result)) )) )
           (verbalise result errmsg counter) ) )
       (define (run-one-verification sv)
         (let* ((name     (car sv))
                (sf       (cadr sv))
                (schecker (caddr sv))
                (n        (cadddr sv)) )
           ;; les champs supplementaires de sv ne sont presents que si n>0
           (if (> n 0)
               (check-one-function sv name sf schecker n)
               (warn "Vous n'avez défini aucun test pour votre fonction"
                     name ) ) ) )
       (for-each (lambda (sv)
                   (if (symbol? (car sv))
                       (run-one-verification sv) ) )
                 student-v ) )

     ;; les fonctions de l'etudiant doivent satisfaire les tests de
     ;; l'enseignant:
     (define (check-teacher-tests student-v teacher-v)
       (define (check-one-function tv name tf tchecker n code sv)
         (define (verbalise result errmsg)
           (cond
            ((and (number? result) (< result 0))
             (let* ((ntests (abs result))
                    (test-expr (fetch-test code ntests)) )
               (record 'teacher-good-tests (- ntests 1))
               (warn-or-abort "Votre fonction" name
                              "échoue sur mon test numéro"
                              ntests (if errmsg errmsg "")
                              (if test-expr
                                  (string-append 
                                   "et voici le test en question: " 
                                   (->string test-expr) )
                                  "" ) ) ) )
            (errmsg
             (warn-or-abort "Votre fonction" name
                            "ne passe pas mes" n 
                            "tests. L'erreur signalée "
                            "est " errmsg "." ) )
            (else
             (record 'teacher-good-tests (- result 1))
             (push "Votre fonction" name
                   "passe tous mes" n "tests" )
             result ) ) )
         (let* ((sf    (cadr sv))
                (names (collect-checked-functions-names tv))
                ;(side0 (debug `(** names ,names)));DEBUG
                (sfns  (collect-named-functions student-v names))
                ;(side1 (debug `(** sfns ,sfns)));DEBUG
                ;; result: (value) + (n errmsg)
                (result (with-handler handler2pair
                                      (list (apply tchecker sfns)) ))
                ;(side2 (debug `(** result ,result)));DEBUG
                ;; errmsg: Exception + #f
                (errmsg (cond ((string? result)
                               result )
                              ((pair? (cdr result))
                               (cadr result) )
                              (else #f) ))
                ;; result: nat (positif si ok, negatif sinon)
                (result (cond ((string? result)
                               -1 )
                              ((pair? (cdr result))
                               (car result) )
                              (else (car result)) )) )
           (verbalise result errmsg) ) )
       (define (run-one-verification tv)
         (let* ((name     (car tv))
                (tf       (cadr tv))
                (tchecker (caddr tv))
                (n        (cadddr tv))
                (code     (let ((rest (cddddr tv)))
                            (if (pair? rest) (car rest) #f) ))
                ;;tobedone
                (sv  (assq name student-v)) )
           (if (> n 0)
               (if (pair? sv)
                   (check-one-function tv name tf tchecker n code sv)
                   (warn-or-abort "Vous n'avez pas défini la fonction"
                                  name ) ) ) ) )
       (push "Je teste maintenant que vos solutions passent mes propres tests")
       (for-each (lambda (tv)
                   (if (symbol? (car tv))
                       (run-one-verification tv) ) )
                 teacher-v )
       (if (not warnings)
           (push "Bravo, vous avez réussi à passer tous mes tests") ) )

     ;;
     (define (verbalize-usages susage tusage usages student-expressions)
       (if (> poids-couverture 0.0)
           ;; seulement si les tests de couverture ont un poids!
           (let ((nsusage (car usages))
                 (ntusage (cadr usages)) )
             (if (> ntusage nsusage)
                 (let ((hints (collect-usage-comparison 
                               susage tusage student-expressions )))
                   (warn "Vos tests utilisent" nsusage 
                         "chemins d'exécution différents"
                         "dans votre code. Mes tests en utilisent" ntusage 
                         ". Il vous manque donc quelques tests. "
                         "Voici quelques indices: " )
                   (for-each warn hints) )
                 (push "Excellent! Vos tests utilisent autant de"
                       "chemins d'exécution différents dans votre"
                       "code que les miens.") ) ) ) )

     (define (verbalise-measures fns tmeasures smeasures)
       (define (extract fname measures)
         (map (lambda (index+counts)
                (cons (car index+counts)
                      (cadr (assq fname (cdr index+counts))) ) )
              measures ) )
       (define (verbalise-comparisons fname scounts tcounts)
         (for-each
          (lambda (i sc tc)
            (cond ((> sc tc)
                   (warn "Pour un problème de taille" i
                         "vous invoquez" fname
                         sc "fois quand je ne l'utilise que"
                         tc "fois." ) )
                  ((= sc tc)
                   (push "Pour un problème de taille" i
                         "vous invoquez" fname
                         "autant de fois que moi.") ) ) )
          (map car scounts)
          (map cdr scounts)
          (map cdr tcounts) ) )
       (define (difference op scounts tcounts)
         (if (pair? scounts)
             (cons (op (car scounts) (car tcounts))
                   (difference op (cdr scounts) (cdr tcounts)) )
             '() ) )
       (define (linear? icounts)
         ;; icounts = ((index1 . value1) ...)
         ;; y = ax + b
         (define (is-linear? icounts)
           ;; on calcule a et b sur les 2 premiers items et on verifie
           ;; avec les suivants.
           (let* ((i1 (caar icounts))
                  (y1 (cdar icounts))
                  (i2 (caadr icounts))
                  (y2 (cdadr icounts))
                  (y* (cddr icounts))
                  (a (/ (- y2 y1) (- i2 i1)))
                  (b (- y1 (* a i1))) )
             (define (check i+v)
               (let ((i (car i+v))
                     (v (cdr i+v)) )
                 (= v (+ (* a i) b)) ) )
             (and (pair? y*)
                  (andmap check y*) ) ) )
         (and  (pair? icounts)
               (pair? (cdr icounts))
               (or (is-linear? icounts)
                   (and (pair? (cddr icounts))
                        (linear? (cdr icounts)) ) ) ) )
       (define (quadratic? icounts)
         ;; icounts = ((index1 . value1) ...)
         ;; y = axx + bx + c
         (define (is-quadratic? icounts)
           (let* ((iy1 (car icounts))
                  (x1 (car iy1))
                  (y1 (cdr iy1))
                  (iy2 (cadr icounts))
                  (x2 (car iy2))
                  (y2 (cdr iy2))
                  (iy3 (caddr icounts))
                  (x3 (car iy3))
                  (y3 (cdr iy3))
                  (y* (cdddr icounts))
                  (a (let* ((dy12 (/ (- y1 y2) (- x1 x2)))
                            (dy13 (/ (- y1 y3) (- x1 3))) )
                       (/ (- dy12 dy13) (- x2 x3)) ))
                  (b (let ((dy12 (/ (- y1 y2) (- x1 x2))))
                       (- dy12 (* a (+ x1 x2))) ))
                  (c (- y1 (* a x1 x1) (* b x1))) )
             (define (check i+v)
               (let ((i (car i+v))
                     (v (cdr i+v)) )
                 (= v (+ (* a i i) (* b i) c)) ) )
             (and (pair? y*)
                  (andmap check y*) ) ) )
         (and (pair? icounts)
              (pair? (cdr icounts))
              (pair? (cddr icounts))
              (or (is-quadratic? icounts)
                  (and (pair? (cdddr icounts))
                       (quadratic? (cdr icounts)) ) ) ) )
       (define (exponential? icounts)
         ;; icounts = ((index1 . value1) ...)
         ;; y = a b^x + c
         #f )
       (define (all op counts)
         (define (allinner first rest)
           (if (pair? rest)
               (and (op first (car rest))
                    (allinner (car rest) (cdr rest)) )
               #t ) )
         (if (pair? counts)
             (allinner (car counts) (cdr counts))
             #f ) )
       (define (estimate-complexity scounts tcounts)
         (define (whatelse malus)
           (let* ((sc (map cdr scounts))
                  (tc (map cdr tcounts))
                  (d (difference - sc tc)) )
             (if (all < d)
                 (begin
                   (record 'malus malus)
                   (warn "La complexité de vos algorithmes"
                         "dépasse la mienne. Je vous inflige un malus"
                         malus "points") ) ) ) )
         (cond ((linear? tcounts)
                (cond ((linear? scounts)
                       (push "Bien! Nous avons tous les deux des algorithmes"
                             "de complexité linéaire.") )
                      ((quadratic? scounts)
                       (warn "Ma solution a une complexité linéaire, la vôtre"
                             "est quadratique! Je vous inflige donc un malus." )
                       (record 'malus 0.15) )
                      (else (whatelse 0.25)) ) )
               ((quadratic? tcounts)
                (cond ((linear? scounts)
                       (push "Bravo! Votre algorithme a une meilleure"
                             "complexité que ma solution, veuillez me le"
                             "signaler afin que j'améliore ma solution" ) )
                      ((quadratic? scounts)
                       (push "Bien! Nous avons tous les deux des algorithmes"
                             "de complexité quadratique.") )
                      (else (whatelse 0.1)) ) )
               (else (whatelse 0.125)) ) )
       (define (verbalise-one-instrumented-function fname)
         (let ((scounts (extract fname smeasures))
               (tcounts (extract fname tmeasures)) )
           (if (equal? scounts tcounts)
               (if (not (andmap zero? (map cdr scounts)))
                   (push "Vous faites autant d'appels à"
                         fname "que moi" ) )
               (begin
                 (estimate-complexity scounts tcounts)
                 (warn "Vous utilisez" fname
                       "différemment de moi." )
                 (verbalise-comparisons fname scounts tcounts) ) ) ) )
       ;; Un marqueur pour specifier le debut des annotations
       ;; concernant la complexite.
       (record 'complexity 'start)
       (for-each verbalise-one-instrumented-function fns) )

     ;; Enchainer tout cela:
     (let* (;; lire sans faute le fichier etudiant:
            (student (read-student-file student-file))
            ;; verifier la syntaxe du fichier etudiant:
            (student-code (check-student-file student))
            ;; collecter les expressions du code etudiant:
            (student-expressions (collect-student-code student-code))
            ;; evaluer le fichier etudiant excluant ses propres tests:
            (student-v (reverse (eval-student-file student-code)))
            (void0 (record 'student-v student-v))
            ;; passer les fonctions de l'etudiant dans ses propres tests
            ;; et verifier que les tests utilisent bien les fonctions:
            (void1 (reset-usage!))
            (ss (check-students-own-tests student-v))
            (susage (reset-usage!)) ; et recuperer profil
            ;; lire sans faute le fichier enseignant:
            (teacher (read-teacher-file teacher-file))
            ;; verifier la syntaxe du fichier enseignant:
            (teacher-code (check-teacher-file teacher))
            ;; evaluer le fichier enseignant, le resultat contient le verifieur
            ;; et le mesureur. Les deux seront a invoquer le moment venu.
            (teacher-v (reverse (eval-teacher-file teacher-code)))
            (debug1 (pp `(teacher-v ,teacher-v))) ; DEBUG
            (void2 (record 'teacher-v teacher-v))
            ;; passer les fonctions de l'enseignant dans ses propres tests:
            (ts (check-teacher-own-tests teacher-v))
            ;; passer les fonctions de l'enseignant dans les tests etudiant:
            ;; ATTENTION! les tests de l'etudiant peuvent exiger plus que
            ;; ce que la specification impose! Ca devrait etre confiné!!!!!
            (st (check-student-tests student-v teacher-v))
            ;; passer les fonctions de l'etudiant dans les tests enseignant:
            (void1 (reset-usage!))
            ;(void2 (reset-profiling!))
            ;(void2 (set! *debug* #t));DEBUG
            (ts (check-teacher-tests student-v teacher-v))
            (tusage (reset-usage!))    ; et recuperer couverture
            ;(hprof (reset-profiling!)) ; et recuperer profil
            ;; comparer la couverture des tests etudiant et enseignant:
            (usages (compare-usage susage tusage student-expressions))
            (void2 (record 'compare-usage usages))
            (void4 (verbalize-usages susage tusage usages student-expressions))
            ;(void3 (record 'profiling hprof))
            (tmeasurer-v (log-find teacher-v adjoined-measurer-tag #f))
            ;(void2b (pp `(tmeasurer-v ,tmeasurer-v)));DEBUG
            (tsmeasures
             (if tmeasurer-v  ; et seulement si tests ok ???
                 (let* ((measure-f (car (cadr tmeasurer-v)))
                        (form      (cadr (cadr tmeasurer-v)))
                        ;; functions to instrument:
                        (fns       (cadr form))
                        ;; functions to invoke:
                        (functions (caddr form))
                        (teacher-measures #f)
                        (student-measures #f) )
                   ;; Alter the global environment
                   (define (set-profiled-functions! fns genv)
                     (for-each (lambda (name)
                                 (let ((f (environment-get name genv)))
                                   (environment-insert!
                                    name
                                    (make-profiling name f)
                                    genv ) ) )
                               fns ) )
                   ;; measure teacher's code:
                   (let ((functions-values (collect-named-functions
                                            teacher-v functions )) )
                     (environment-set! teacher-env)
                     (set-profiled-functions! fns teacher-env)
                     (let ((measures (apply measure-f functions-values)))
                       (record 'teacher-profiling measures)
                       (set! teacher-measures measures) ) )
                   ;; measure student's code:
                   (let ((functions-values (collect-named-functions
                                            student-v functions )) )
                     (environment-set! student-env)
                     (set-profiled-functions! fns student-env)
                     (let ((measures (apply measure-f functions-values)))
                       (record 'student-profiling measures)
                       (set! student-measures measures) ) )
                   (verbalise-measures fns teacher-measures student-measures)
                   'done )
                 ;; absent (li101-verify ...) form:
                 #f ) )
            (voidend 'done) )
       (reverse log) ) ) ) )

;;; Extraction du texte d'un test. Les tests sont numerotes a partir de 1.
(define (fetch-test code i)
  ;; find code for ith test.
  ;; <code> = (if (...) <code> (li101- <i> (quote <test>)))
  ;; <code> = (if (...) <code> (li101- (cons <i> result) (quote <test>)))
  ;; <code> = n     ; nombre total de tests + 1
  (define (fetch code i)
    (if (pair? code)
        (cond ((eq? 'if (car code))
               (fetch-from-if code i) )
              ((eq? 'let (car code))
               (fetch (caddr code) i) )
              ((is-verify? code)
               (fetch-from-verify (cddr code) i) )
              (else #f) )
        #f ) )
  ;; <code> = (exp1 => valeur1 ...)
  (define (fetch-from-verify clauses i)
    (if (pair? clauses)
        (let ((exp1 (car clauses)))
          (if (<= i 1)
              exp1 
              (fetch-from-verify (cdddr clauses) (- i 1)) ) )
        #f ) )
  (define (fetch-from-if code i)
    ;; assume code ~ (if ...)
    (if (and (pair? (cdr code))
             (pair? (cddr code))
             ; code ~ (if ? <code> ...)
             ;(begin (debug `(code ,(caddr code))) #t)
             (pair? (cdddr code))
             (pair? (cadddr code))
             ;(begin (debug `(====> ,(cadddr code))) #t)
             (eq? 'li101- (car (cadddr code)))
             (pair? (cdr (cadddr code)))
             (pair? (cddr (cadddr code)))
             ; code ~ (if ? <code> (li101- ? ??))
             )
        (let* ((code+test (cddr code))
               (code (car code+test))
               ;; code ~ nat | (cons nat result)
               (j (if (pair? (cadr (cadr code+test)))
                      (cadr (cadr (cadr code+test)))
                      (cadr (cadr code+test)) )) )
          (cond ((> i j) (fetch code i))
                ((= i j) (cadr (caddr (cadr code+test))))
                (else #f) ) )
        #f ) )
  (if code (fetch code i) #f) )

;;; Version simpliste pour exercice graphique. On veut juste evaluer
;;; le fichier de l'etudiant, la valeur finale doit etre l'image
;;; attendue.
(define (eval-without-compare student-file . environment)
  ;; teacher-file is not used!
  (define env (if (pair? environment)
                  (car environment)
                  *the-environment* ))
  (define log (list))
  (reset-profiling!)
  (reset-usage!)
  (call/cc 
   (lambda (exit)
     (define (->single-string msgs)
       (apply string-append 
              (map (lambda (o) 
                     (string-append " " (->string o)) )
                   msgs ) ) )
     (define (abort . msgs)
       (let ((s (->single-string msgs)))
         (debug `(*** error ,s))
         (set! log (cons (list 'error s) log))
         (exit (reverse log)) ) )
     (define (record kind value)
       (debug `(*** data ,kind ,value))
       (set! log (cons (list 'data kind value) log)) )
     (define (push . msgs)
       (let ((s (->single-string msgs)))
         (debug `(*** string ,s))
         (set! log (cons (list 'string s) log)) ) )
     (define (warn . msgs)
       (let ((s (->single-string msgs)))
         (debug `(*** warning ,s))
         (set! log (cons (list 'warning s) log)) ) )

     ;; Chercher dans v* les valeurs des fonctions nommees dans le
     ;; champ fns. Substituer fname a la fonction name. Retourner la
     ;; liste de ces fonctions.
     (define (collect-checked-functions v* v name fname)
       (let ((fns (collect-checked-functions-names v)))
         (debug `(fns ,fns));DEBUG
         (map (lambda (othername)
                (if (eq? name othername)
                    fname
                    (let ((otherv (assq othername v*)))
                      (if (pair? otherv)
                          (cadr otherv)
                          (li101-error 1134
                                       "Je ne trouve pas votre définition de"
                                       "la fonction" othername) ) ) ) )
              fns ) ) )
     (define (collect-checked-functions-names v)
       (caddr (cdddr v)) )

     ;; lire le fichier de l'étudiant:
     (define (read-student-file student-file)
       (push "Je vais lire votre fichier")
       (let ((student (with-handler handler2string
                                    (slurp-file student-file) )))
         (if (string? student)
             (abort "Votre fichier ne peut être lu (" student ")")
             student ) ) )

     ;; le vérifier syntaxiquement:
     (define (check-student-file student)
       (push "J'analyse la syntaxe de votre fichier")
       (let ((student-code (with-handler handler2string
                                         (check-program? student) )))
         (if (string? student-code)
             (abort "Votre fichier est syntaxiquement incorrect ("
                    student-code ")" )
             student ) ) )

     ;; l'évaluer. Si le fichier s'evalue correctement. Attention, les
     ;; tests de l'etudiant ne sont pas encore executes.
     (define (eval-student-file student-code)
       (push "J'évalue votre fichier")
       (let ((student-v (with-handler handler2string
                                      (li101-program-eval 
                                       student-code
                                       (list)
                                       env ) )))
         (if (string? student-v)
             ;; Dans ce cas stt sera nul.
             (abort "Votre fichier s'évalue erronément (" student-v ")")
             (let* ((funsv (filter (lambda (v)
                                     (symbol? (car v)) )
                                   student-v ))
                    (names (map car funsv))
                    (ntests (apply + (map cadddr funsv))) )
               (cond
                ((= 1 (length names))
                 (push "Vous avez défini la fonction " (car names)) )
                (else
                 (apply push "Vous avez défini les fonctions suivantes "
                        (reverse names) ) ) )
               ;; enregistrer le nombre total de tests ecrits par l'etudiant:
               (record 'student-total-tests ntests) ) )
         student-v ) )

     ;; On passe les fonctions de l'etudiant dans ses test afin de
     ;; mesurer son usage. C'est la meme fonction que dans eval-and-compare
     (define (check-students-own-tests student-v)
       (define (check-one-function sv name sf schecker n)
         (define (verbalise result errmsg counter)
           (cond
            ((and (number? result) (< result 0))
             (let ((ntests (- (abs result) 1)))
               ;; nombre de tests reussis:
               (record 'student-good-own-tests ntests)
               (cond
                ((= ntests 0)
                 (abort "Votre fonction" name
                        "ne passe aucun de vos tests"
                        (if errmsg errmsg "") ) )
                ((= ntests 1)
                 (abort "Votre fonction " name "ne passe que le"
                        "premier de vos tests"
                        (if errmsg errmsg "") ) )
                (else
                 (abort "Votre fonction" name
                        "échoue sur votre test numéro"
                        (+ ntests 1) "je m'arrête là"
                        (if errmsg errmsg "") ) ) ) ) )
            (errmsg
             (abort "Votre fonction" name "est erronée et"
                    "ne passe pas vos propres" n "tests ("
                    errmsg ")" ) )
            ((= counter 0)
             (record 'student-good-own-tests (- result 1))
             (abort "Vos tests n'utilisent pas votre fonction"
                    name ) )
            (else
             (record 'student-good-own-tests (- result 1))
             (push "Votre fonction" name
                   "passe tous vos" n "tests" )
             result ) ) )
         (let* ((counter 0)
                (scountingf (lambda args
                              (set! counter (+ 1 counter))
                              (apply sf args) ))
                (names (collect-checked-functions-names sv))
                (fns   (collect-checked-functions
                        student-v sv name scountingf ))
                ;; result: (value) + (n errmsg)
                (result (with-handler handler2pair
                                      (list (apply schecker fns)) ))
                ;; errmsg: Exception + #f
                (errmsg (cond ((string? result)
                               result )
                              ((pair? (cdr result))
                               (cadr result) )
                              (else #f) ))
                ;; result: nat (positif si ok, negatif sinon)
                (result (cond ((string? result)
                               -1 )
                              ((pair? (cdr result))
                               (car result) )
                              (else (car result)) )) )
           (verbalise result errmsg counter) ) )
       (define (run-one-verification sv)
         (let* ((name     (car sv))
                (sf       (cadr sv))
                (schecker (caddr sv))
                (n        (cadddr sv)) )
           ;; les champs supplementaires de sv ne sont presents que si n>0
           (if (> n 0)
               (check-one-function sv name sf schecker n)
               (warn "Vous n'avez défini aucun test pour votre fonction"
                     name ) ) ) )
       (for-each (lambda (sv)
                   (if (symbol? (car sv))
                       (run-one-verification sv) ) )
                 student-v ) )

     ;; Enchainer tout cela:
     (let* (;; lire sans faute le fichier etudiant:
            (student (if (string? student-file)
                         (read-student-file student-file)
                         ;; HACK pour tests:
                         student-file ))
            ;; verifier la syntaxe du fichier etudiant:
            (student-code (check-student-file student))
            ;; evaluer le fichier etudiant excluant ses propres tests:
            (student-v (reverse (eval-student-file student-code)))
            (void0 (record 'student-v student-v))
            ;; passer les fonctions de l'etudiant dans ses propres tests
            ;; et verifier que les tests utilisent bien les fonctions:
            (void1 (reset-usage!))
            (ss (check-students-own-tests student-v)) )
       (reverse log) ) ) ) )

;;; Convert the log into some XML fragment to be shown to the student:

(define (shape-normal s)
  (string-append "<li>&#x2714; " s "</li>
") )
(define (shape-error s)
  (string-append "<li><error>&#x2718; " s "</error></li>
") )
(define (shape-warning s)
  (string-append "<li>&#x2718; " s "</li>
") )

;;; Extract a slice of the log between (data start ...) and
;;; (data stop ...) excluded. The log may be seen as three XML fragments:
;;; the first till compare-usage is for teacher's tests
;;; the second till complexity is for test coverage
;;; the third starting at complexity is for complexity
(define (log-split log data-sentinel-start data-sentinel-stop)
  (define (copy log)
    (if (pair? log)
        (let* ((line (car log))
               (kind (car line)) )
          (if (and (eq? kind 'data)
                   data-sentinel-stop
                   (eq? (cadr line) data-sentinel-stop) )
              '()
              (cons line (copy (cdr log))) ) )
        '() ) )
  (define (skip log)
    (if (pair? log)
        (let* ((line (car log))
               (kind (car line)) )
          (if (and (eq? kind 'data)
                   data-sentinel-start
                   (eq? (cadr line) data-sentinel-start) )
              (copy (cdr log))
              (skip (cdr log)) ) )
        '() ) )
  (if data-sentinel-start
      (skip log)
      (copy log) ) )

(define (shape-line line)
  (let ((kind (car line)))
    (case kind
      ((string)  (shape-normal (cadr line)))
      ((error)   (shape-error (cadr line)))
      ((warning) (shape-warning (cadr line)))
      (else      "") ) ) )

(define (log2xml log data-sentinel-start data-sentinel-stop)
  (let ((strings (map shape-line
                      (log-split log data-sentinel-start data-sentinel-stop) )))
    (if (pair? strings)
        (string-append "<ul>"
                       (apply string-append strings)
                       "</ul>" )
        "" ) ) )
  
(define (log2xml1 log)
  (log2xml log #f 'compare-usage) )

(define (log2xml2 log)
  (log2xml log 'compare-usage 'complexity) )

(define (log2xml3 log)
  (log2xml log 'complexity #f) )
    
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (collect-usage-comparison susage tusage fragments)
  ;; ?usage are hashtable as returned by reset-usage!
  (let ((exps (hashtable-key-list tusage))
        (results (list)) )
    (define (generate-msg exp tindex)
      (let* ((limit 800)
             (exp (if (and (pair? exp) (null? (cdr exp)))
                      (car exp)
                      exp ))
             (expstring (->string exp))
             (expstring (if (> (string-length expstring) limit)
                            (string-append (substring expstring 0 limit) "...")
                            expstring )) )
        (if (>= tindex 99999)
            ;; Entree dans une fonction utilisateur:
            (string-append "Vos tests n'ont pas activé la fonction dont " 
                           "le corps débute par " expstring )
            ;; tindex-eme (zero-based) cas de sortie:
            (cond ((and (pair? exp)
                        (eq? (car exp) 'if) )
                   (string-append 
                    "Vos tests n'ont pas activé " 
                    (if (= 0 tindex) "la conséquence" "l'alternant")
                    " de votre alternative "
                    expstring ) )
                  ((and (pair? exp)
                        (eq? (car exp) 'cond) )
                   (string-append 
                    "Vos tests n'ont pas activé le " 
                    (if (= 0 tindex) "premier" 
                        (string-append (->string (+ 1 tindex)) "ème") )
                    " cas de votre conditionnelle "
                    expstring ) )
                  ((and (pair? exp)
                        (memq (car exp) '( and or )) )
                   (string-append 
                    "Vos tests n'ont pas activé le " 
                    (if (= 0 tindex) "premier" 
                        (string-append (->string (+ 1 tindex)) "ème") )
                    " cas de votre expression "
                    expstring ) )
                  (else 
                   (string-append 
                    "Vos tests n'ont pas activé le " 
                    (->string (+ 1 tindex))
                    (if (= tindex 0) "er" "ème")
                    " cas de " 
                    expstring ) ) ) ) ) )
    (for-each 
     (lambda (exp)
       (if (memq exp fragments)
           (let ((talist (hashtable-get tusage exp)))
             (for-each
              (lambda (ti+c)
                (let ((tindex (car ti+c))
                      (tcount (cdr ti+c)) )
                  (if (> tcount 0)
                      (let ((scount (get-usage susage exp tindex)))
                        (if (= scount 0)
                            (set! results
                                  (cons (generate-msg exp tindex)
                                        results ) ) ) ) ) ) )
              (sort (lambda (x y)
                      (<= (car x) (car y)) )
                    talist ) ) ) ) )
     exps )
    (reverse results) ) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Compute a mark!

;;; compute the first part of the mark on the student's code:
(define (compute-mark1 details)
  (let ((stt 0)              ; numberOfTestsWrittenByStudent
        (ttt 0)              ; numberOfTestsWrittenByTeacher
        (sgt 0)              ; numberOfStudentsTestsSucceededByTeacher
        (tgt 0)              ; numberOfTeachersTestsSucceededByStudent
        (scover 0)           ; numberOfCodePathsUsedByStudentsTests
        (tcover 0) )         ; numberOfCodePathsUsedByTeachersTests
    (set! stt (cadr details))
    (set! details (cddr details))
    (set! ttt (cadr details))
    (set! details (cddr details))
    (set! sgt (cadr details))
    (set! details (cddr details))
    (set! tgt (cadr details))
    (set! details (cddr details))
    (set! scover (cadr details))
    (set! details (cddr details))
    (set! tcover (cadr details))
    (cond 
     ;; problème avant même de pouvoir compter les tests de l'enseignant. 
     ;; il est interdit a l'enseignant de n'ecrire aucun test!
     ((= ttt 0) 0.0)
     ;; le programme de l'étudiant ne passe pas ses propres tests ou
     ;; alors l'etudiant n'a ecrit aucun test (ce qui est le cas dans
     ;; l'examen syracuse).
     ((= stt 0)
      (let* ((correction2 (/ tgt ttt)))
        (* (+ poids-correction1 poids-correction2) correction2) ) )
     ;; Cas normal:
     (else (let ((correction1 (/ sgt stt))
                 (correction2 (/ tgt ttt)) )
             (+ (* poids-correction1 correction1)
                (* poids-correction2 correction2) ) )) ) ) )

;;; compute the second part of the mark on the student's tests
(define (compute-mark2 details)
  (let ((stt 0)              ; numberOfTestsWrittenByStudent
        (ttt 0)              ; numberOfTestsWrittenByTeacher
        (sgt 0)              ; numberOfStudentsTestsSucceededByTeacher
        (tgt 0)              ; numberOfTeachersTestsSucceededByStudent
        (scover 0)           ; numberOfCodePathsUsedByStudentsTests
        (tcover 0) )         ; numberOfCodePathsUsedByTeachersTests
    (set! stt (cadr details))
    (set! details (cddr details))
    (set! ttt (cadr details))
    (set! details (cddr details))
    (set! sgt (cadr details))
    (set! details (cddr details))
    (set! tgt (cadr details))
    (set! details (cddr details))
    (set! scover (cadr details))
    (set! details (cddr details))
    (set! tcover (cadr details))
    (cond 
     ;; le programme de l'étudiant ne passe pas ses propres tests:
     ((and (> poids-couverture 0.0) (= stt 0)) 0.0)
     ;; problème avant même de pouvoir compter les tests de l'enseignant:
     ((= ttt 0) 0.0)
     ;; Cas normal:
     (else (let* ((couverture (if (= tcover 0) 
                                  0.0
                                  (/ scover tcover) )))
             (* poids-couverture couverture) )) ) ) )

;;; compute the third part of the mark: the malus
(define (compute-mark3 details)
  (let ((stt 0)              ; numberOfTestsWrittenByStudent
        (ttt 0)              ; numberOfTestsWrittenByTeacher
        (sgt 0)              ; numberOfStudentsTestsSucceededByTeacher
        (tgt 0)              ; numberOfTeachersTestsSucceededByStudent
        (scover 0)           ; numberOfCodePathsUsedByStudentsTests
        (tcover 0)           ; numberOfCodePathsUsedByTeachersTests
        (malus 0) )          ; malus
    (set! stt (cadr details))
    (set! details (cddr details))
    (set! ttt (cadr details))
    (set! details (cddr details))
    (set! sgt (cadr details))
    (set! details (cddr details))
    (set! tgt (cadr details))
    (set! details (cddr details))
    (set! scover (cadr details))
    (set! details (cddr details))
    (set! tcover (cadr details))
    (set! details (cddr details))
    (set! malus (cadr details))
    malus ) )

;;; 

(define poids-correction1 0.)
(define poids-correction2 0.5)
(define poids-couverture  0.5)

(define (do-compute-mark stt ttt sgt tgt scover tcover)
  (cond 
   ;; le programme de l'étudiant ne passe pas ses propres tests:
   ((= stt 0) 0.0)
   ;; problème avant même de pouvoir compter les tests de l'enseignant:
   ((= ttt 0) 0.0)
   ;; Cas normal:
   (else (let* ((correction1 (/ sgt stt))
                (correction2 (/ tgt ttt))
                (couverture  (if (= tcover 0)
                                 (/ (+ correction1 correction2) 2)
                                 (/ scover tcover) )) )
           (+ (* poids-correction1 correction1)
              (* poids-correction2 correction2)
              (* poids-couverture couverture) ) )) ) )

(define (compute-mark-details log)
  (let ((stt 0)     ; numberOfTestsWrittenByStudent
        (ttt 0)     ; numberOfTestsWrittenByTeacher
        (sgt 0)     ; numberOfStudentsTestsSucceededByTeacher
        (tgt 0)     ; numberOfTeachersTestsSucceededByStudent
        ;; These are measured on student's code:
        (scover 0)  ; number of code paths used by student's tests
        (tcover 0)  ; number of code paths used by teacher's tests
        ;; and maybe some malus
        (malus 0) )
    (for-each
     (lambda (line)
       (if (eq? (car line) 'data)
           (case (cadr line)
             ((student-total-tests)
              ;; Le nombre total de tests qu'a ecrits l'etudiant:
              (set! stt (+ stt (caddr line))) )
             ((teacher-total-tests) 
              ;; Le nombre total de tests qu'a ecrits l'enseignant:
              (set! ttt (+ ttt (caddr line))) )
             ((student-good-tests)  
              ;; Le nombre total de tests ecrits par l'etudiant qu'a
              ;; reussis l'enseignant donc sgt <= stt
              (let ((n (abs (caddr line))))
                (set! sgt (+ sgt n)) ) )
             ((teacher-good-tests)  
              ;; Le nombre total de tests ecrits par l'enseignant qu'a
              ;; reussis l'etudiant donc tgt <= ttt
              (let ((n (abs (caddr line))))
                (set! tgt (+ tgt n)) ) )
             ((compare-usage)
              (let ((s+t (caddr line)))
                (set! scover (car s+t))
                (set! tcover (cadr s+t)) ) )
             ((malus)
              (set! malus (+ malus (caddr line))) )
             (else 0) ) ) )
     log )
    (debug `(*** stt ,stt ttt ,ttt sgt ,sgt tgt ,tgt 
                 scover ,scover tcover ,tcover
                 malus ,malus ))
    (list "numberOfTestsWrittenByStudent" stt
          "numberOfTestsWrittenByTeacher" ttt
          "numberOfStudentsTestsSucceededByTeacher" sgt
          "numberOfTeachersTestsSucceededByStudent" tgt
          "numberOfCodePathsUsedByStudentsTests" scover 
          "numberOfCodePathsUsedByTeachersTests" tcover
          "malus" malus
          ) ) )

;;; Display profiling
(define (display-profiling out)
  (hashtable-for-each 
   hprofiling 
   (lambda (name n)
     (display `(,name used ,n times) out)
     (newline out) ) ) )

;;; }}}
;;; {{{ Helper for tests
;;; These functions should be present in build-primitive-environment

(define (approx= r1 r2)
  (< (abs (- r1 r2)) 0.1) )

(define (set= set1 set2) 
  (if (pair? set1)
      (and (member (car set1) set2)
           (set= (cdr set1) (remove (car set1) set2)) )
      (null? set2) ) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Enchainer tout dans li101.sh
;;; Le post-traitement peut etre modifie dans li101+.scm

(define (li101-postprocess log details)
  log )

;;; Find a record in the log or return #f.
(define (log-find log key subkey)
  (if (pair? log)
      (let ((v (car log)))
        (debug `(** log-find analyze ,v ,key ,subkey));DEBUG
        (if (and (pair? v)
                 (eq? key (car v))
                 (if (and subkey (pair? (cdr v)))
                     (eq? subkey (cadr v))
                     #t ) )
            v ; the whole record!
            (log-find (cdr log) key subkey) ) )
      #f ) )

;;; find the last value of a program
(define (sv-find-last-value sv)
  (if (pair? sv)
      (if (eq? (caar sv) adjoined-value-tag)
          (cadr (car sv))
          (sv-find-last-value (cdr sv)) )
      #f ) )

(define (li101-generate-output-files log tmpdir)
  (call-with-output-file (string-append tmpdir "/li101a.out")
    (lambda (out) (display (log2xml1 log) out)) )
  (if (> poids-couverture 0.0)
      (call-with-output-file (string-append tmpdir "/li101b.out")
        (lambda (out) (display (log2xml2 log) out)) ) )
  (call-with-output-file (string-append tmpdir "/li101.profiling")
    (lambda (out) (display-profiling out)) )
  (let ((details (compute-mark-details log)))
    (call-with-output-file (string-append tmpdir "/li101.details")
      (lambda (out) (display details out)) )
    ;;(call-with-output-file (string-append tmpdir "/li101.mark"); note totale
    ;;  (lambda (out) (display (compute-mark log) out)) )
    (call-with-output-file (string-append tmpdir "/li101a.mark") ; note code
      (lambda (out) (display (compute-mark1 details) out)) )
    (if (> poids-couverture 0.0)
        (call-with-output-file (string-append tmpdir "/li101b.mark") ; note tests
          (lambda (out) (display (compute-mark2 details) out)) ) )
    (let ((malus (compute-mark3 details)))
      (call-with-output-file (string-append tmpdir "/li101c.out")
        (lambda (out) (display (log2xml3 log) out)) )
      (if (> malus 0)
          (call-with-output-file (string-append tmpdir "/li101c.mark") ; malus
            (lambda (out) (display (- malus) out)) ) ) )
    details ) )

;;; Exercice normal.
(define (li101-process student-file teacher-file environment tmpdir)
  (define (logger revlog)
    ;; Generates the output files whenever a new piece of log is
    ;; generated. This allows to have the partial results even if the
    ;; program is killed by lack of time or cpu.
    (li101-generate-output-files (reverse revlog) tmpdir) )
  (define (process)
    (let ((log (eval-and-compare student-file teacher-file environment logger)))
      ;; will appear in /li101.err (see li101.sh) and will only appear
      ;; in the grading report if FW4EX_SHOW_BIGLOO_OUTPUT is true.
      ;(display "===============before display log")
      (display log) 
      ;(display "===============after process")
      (let ((details (li101-generate-output-files log tmpdir)))
        ;(display "===============entering post-process")
        ;; Post-traitement possible:
        (li101-postprocess log details) ) ) )
  (let ((result (with-handler handler2string 
                     (process) )))
    (if (string? result)
        (call-with-output-file (string-append tmpdir "/li101-.err")
          (lambda (out) (write result out)) ) )
    result ) )

;;; Exercice menant à une image
;;; La dernière expression doit produire un dessin (du .ps) qui sera
;;; convertie en .png et affichée dans le rapport de correction.
;;; Placer un fichier li101+.scm dans l'exercice avec:
;;;       ;;; Exercice graphique
;;;       (set! li101-process li101-process-image)
(define (li101-process-image student-file teacher-file environment tmpdir)
  (define (process)
    (let ((log (eval-without-compare student-file environment)))
      (display log)
      (let* ((sv (caddr (log-find log 'data 'student-v)))
             (last-value (sv-find-last-value sv)) )
        (call-with-output-file (string-append tmpdir "/li101img.eps")
          (lambda (out) (display (image2postscript last-value) out)) ) )
      (call-with-output-file (string-append tmpdir "/li101a.out")
        (lambda (out) (display (log2xml1 log) out)) )
      (let ((details (compute-mark-details log)))
        (call-with-output-file (string-append tmpdir "/li101.details")
          (lambda (out) (display details out)) )
        ;;(call-with-output-file (string-append tmpdir "/li101.mark"); note totale
        ;;  (lambda (out) (display (compute-mark log) out)) )
        (call-with-output-file (string-append tmpdir "/li101a.mark") ; note code
          (lambda (out) (display 1.0 out)) )
        ;; Post-traitement possible:
        (li101-postprocess log details) ) ) )
  (let ((result (with-handler handler2string 
                     (process) )))
    (if (string? result)
        (call-with-output-file (string-append tmpdir "/li101-.err")
          (lambda (out) (display result out)) ) )
    result ) )

;;; Exercice sans correction. C'est utile pour demoeval
;;; Placer un fichier li101+.scm dans l'exercice avec:
;;;       ;;; Exercice sans correction
;;;       (set! li101-process li101-no-process)
(define (li101-no-process student-file teacher-file environment tmpdir)
  (define (process)
    (let ((log (eval-without-compare student-file environment)))
      (display log)
      (call-with-output-file (string-append tmpdir "/li101a.out")
        (lambda (out) (display "<p>Cet exercice n'est pas corrigé!</p>" out)) )
      (call-with-output-file (string-append tmpdir "/li101a.mark") ; note code
        (lambda (out) (display 1.0 out)) )
      ;; Post-traitement possible:
      (li101-postprocess log details) ) )
  (let ((result (with-handler handler2string 
                     (process) )))
    (if (string? result)
        (call-with-output-file (string-append tmpdir "/li101-.err")
          (lambda (out) (display result out)) ) )
    result ) )

;;; }}}
;;; end of li101.scm
